import os
import re
import sys
import string

import arcpy
from code_library import isiterable
from code_library.common.geospatial import generate_fast_filename

import log
import local_vars
import funcs


connectivity_index = {}
network_end_hucs = ["CLOSED BASIN", "Mexico", "OCEAN"]
diversity_network = None


class empty_object:
	""" @todo: this should be replaced in usage with object()"""

	def __init__(self):
		pass


class watershed():
	def __init__(self):
		self.HUC_12 = None
		self.assemblage = []
		self.downstream = None
		self.upstream = []
		self.has_dam = False
		self.beta_div = None


class function_arg:
	''' a small class that allows us to pass around function names with an argument (or set of arguments) in a way that lets us test for it'''

	def __init__(self, function, argument):
		self.function = function
		self.argument = argument  # should be a tuple

		try:
			self.function_obj = globals()[self.function]
		except:
			log.error("function used in function_arg in callback does not exist")
			raise


class query():
	"""
		An ORM would do many of the things that this class does, but we don't have one, so this will have to work
	"""

	def __init__(self):
		self.tables = []
		self.base = None
		self.where_clauses = []
		self.fish_only = True
		self.observation_types = None
		self.suffix = None

	def __str__(self):
		return self.compose()

	def __unicode__(self):
		return self.compose()

	def _compose_simple_query(self):
		sql_tables_str = " from"
		for table in self.tables:
			sql_tables_str += " %s," % table
		sql_tables_str = sql_tables_str[:-1]  # chop off the last ,

		sql_where_str = " where"
		for where in self.where_clauses:
			sql_where_str += " %s and" % where
		sql_where_str = sql_where_str[:-3]  # chop off the last "and"

		if self.suffix:
			str_suffix = self.suffix
		else:
			str_suffix = ""

		return self.base + sql_tables_str + sql_where_str + str_suffix

	def compose(self):
		"""

		"""
		if not self.base:
			raise ValueError("You must define query.base as the 'select' (or other) portion of the sql query")
		if not self.observation_types:
			log.write("Warning: No observation_types set. Will output all types together", 1)

		if not "distinct" in self.base:
			log.write(
				"Warning: 'distinct' is not present in your database query. This doesn't mean it's incorrect, but most"
				"summary queries should use distinct to ensure non-duplication of species", 1)

		if self.observation_types:
			self.where_clauses.append("Observations.Presence_Type in (%s)" % self.observation_types)

		self.where_clauses = list(set(self.where_clauses)) # dedupe it!
		self.tables = list(set(self.tables)) # dedupe it too - required

		return self._compose_simple_query()

	def set_defaults(self, zonal_callback=True, qc=False):
		"""
			Handles the proper method chaining for most queries
		"""
		self.add_base()
		self.set_current()

		if zonal_callback:
			self.add_zonal_callback()

		if qc is True:
			self.make_into_qc()

	def add_base(self):
		"""

		"""

		self.tables.append(local_vars.observations_table)

		if self.fish_only:
			self.tables.append("%s as Species" % local_vars.fish_species_table)
			self.species_table = local_vars.fish_species_table
		else:
			self.tables.append(local_vars.species_table)
			self.species_table = local_vars.species_table

		self.where_clauses.append("%s.Species_ID = %s.FID" % (local_vars.observations_table, local_vars.species_table))

	def add_zonal_callback(self):
		self.where_clauses.append("Observations.Zone_Id = ?")

	def make_count_query(self):
		self.base = "SELECT Count(*) AS col_value FROM ("
		self.suffix = ")"

	def make_into_qc(self):
		"""

		"""
		self.tables.append("Observation_Collections")
		self.collections = local_vars.hq_collections
		self.where_clauses.append("Observation_Collections.Collection_ID in (%s)" % local_vars.hq_collections)
		self.where_clauses.append(
			"%s.OBJECTID = Observation_Collections.Observation_ID" % local_vars.observations_table)

	def set_current(self):
		self.observation_types = local_vars.current_obs_types

	def set_historic(self):
		self.observation_types = local_vars.historic_obs_types


def get_arg(args, index, default, fieldmap=None, split_commas=False):
	try:
		if args[index]:
			if fieldmap and args[index] in fieldmap: # if we want to transform this to something else
				return fieldmap[args[index]]
			if split_commas:
				return split(args[index], ",")
			else:
				return args[index] # otherwise, return what was passed in
		else: # no value, return default
			if fieldmap and default in fieldmap: # if we want to transform this to something else
				return fieldmap[default]
			return default
	except:  # error, return the default. Error should be in getting args[index] - doesn't exist
		if fieldmap and default in fieldmap: # if we want to transform this to something else
			return fieldmap[default]
		return default


def mega_diversity_info(zones_layer, db_cursor, args, parent_layer):
	"""
		runs a large number of diversity functions on a single layer and returns it. These functions all add columns with data to one layer!

		Metadata: In order

		Note that since this function runs a number of other mapping functions, there will be some data duplication.
		Many of the non-qced output columns can now be deprecated since all data of interest is currently qced. They are here for legacy purposes and for any future changes

		-------- BASIC RICHNESS --------
		native_richness: integer. Raw alpha richness as count of number of native fish species in the HUC across PISCES.
		native_qc_richness: integer. Same as above, but only for species in a data collection that is qced
		native_qc_historic_richness: integer.  Same as native_qc_richness, but only for historic data types

		nonnative_richness: integer. Alpha richness of nonnatives. As of this writing, very little data on this.
		nonnative_qc_richness: integer. same as previous, but for data in qc dataset. Currently none for nonnatives, but here for future-proofing.

		all_richness: integer. Alpha richness of all fish species (native or nonnative) in PISCES, does not include amphibs and reptiles, etc.
		all_qc_richness: integer. Same as above, but for qced data. Should match native_qc_richness as of this writing.

		sens_nat_rich: integer. Richness of sensitive native species. Alpha richness of native fish taxa in PISCES where the Moyle 2011 average score <= 3 and >= 0. Uses the field Species_Aux.Average in PISCES for the score.
		sens_nat_qc_rich: integer. Same as previous except only for qced data in each HUC12

		sens_nat_assem: Assemblage of sensitive species in the HUC12 where sensitive is defined as in sens_nat_rich
		sens_nat_qc_assem: same as sens_nat_assem, but only for qc taxa

		-------- SENSITIVITY STATS --------
		avg_sensitivity_score: real (decimal). The average score (Species_Aux.Average) for all taxa in the HUC12
		avg_sensitivity_under3: real (decimal). same as above, but only including species with scores <= 3 and >= 0
		avg_sensitivity_under2: real (decimal). same as above but <=2 and >=0
		avg_sensitivity_under1: real (decimal). same as above but <=1 and >=0

		avg_qc_sensitivity_score: real (decimal). The average score (Species_Aux.Average) for all QCed taxa in the HUC12
		avg_qc_sensitivity_under3: real (decimal). same but <=3 and >=0
		avg_qc_sensitivity_under2: real (decimal). same but <=2 and >=0
		avg_qc_sensitivity_under1: real (decimal). same but <=1 and >=0

		-------- LOSSES AND EXTIRPATIONS --------
		All data in this section is related to native fish and qced data

		current_richness: integer. QCed alpha richness for native fish species. Includes translocations
		current_assemblage: Assemblage of qced native fish species in the HUC12. Includes translocations

		historic_richness: integer. QCed *historic* alpha richness for native fish species
		historic_assemblage: Assemblage of qced native fish species historically in the HUC12

		losses: integer. Effectively, historic_richness - current_richess, except excludes translocations from current. Essentially, a count of extirpated species
		losses_list: The assemblage of species extirpated from that HUC12.
		gains: integer. How many native fish taxa have been translocated into that HUC12?
		gains_list. The assemblage of translocated species

		richness_difference: integer. count of the number of species gained-lost (so # currently - # historically). When the layer is exported without translocations included, this # may not correctly include translocations.

		-------- JACCARD DISTANCES --------
		To be updated later. Please ignore these for now

		-------- DOWNSTREAM INFO --------
		downstream_assemblage: Assemblage of species in the HUC12 downstream

	@param zones_layer:
	@param db_cursor:
	@param args:
	@param parent_layer:
	@return:
	"""

	# skip the passed in cursor - we'll make our own

	run_non_qc = get_arg(args, 0, "0")

	args_native = [
		'SELECT Count(*) AS col_value FROM (select distinct Observations.Species_ID from Observations, Fish_Species '
		'where Observations.Zone_Id = ? And Observations.Species_ID = Fish_Species.FID and Fish_Species.Native = True'
		' and Observations.Presence_Type in (%s))' % (
		local_vars.current_obs_types), 'native_richness', 'LONG']

	args_native_qc = [
		'SELECT Count(*) AS col_value FROM '
		'(select distinct Observations.Species_ID from Observations, Fish_Species, Observation_Collections'
		' where Observations.Zone_Id = ? And Observations.Species_ID = Fish_Species.FID and Fish_Species.Native = True'
		' and Observations.Presence_Type in (%s) and Observations.OBJECTID = Observation_Collections.Observation_ID'
		' and Observation_Collections.Collection_ID in (%s))' % (
		local_vars.current_obs_types, local_vars.hq_collections), 'native_qc_richness', 'LONG']

	args_native_qc_historic = [
		'SELECT Count(*) AS col_value FROM '
		'(select distinct Observations.Species_ID from Observations, Fish_Species, Observation_Collections'
		' where Observations.Zone_Id = ? And Observations.Species_ID = Fish_Species.FID and Fish_Species.Native = True'
		' and Observations.Presence_Type in (%s) and Observations.OBJECTID = Observation_Collections.Observation_ID'
		' and Observation_Collections.Collection_ID in (%s))' % (
		local_vars.historic_obs_types, local_vars.hq_collections), 'native_qc_historic_richness', 'LONG']

	db_connection, db_cursor = funcs.refresh_cursor(database=local_vars.maindb)
	if run_non_qc == "1" or run_non_qc == 1:
		log.write("Running Native Richness", 1)
		t_layer = postprocess_zones(zones_layer, db_cursor, args_native, parent_layer)
	else:
		t_layer = zones_layer

	log.write("Running QC Native Richness", 1)
	db_connection, db_cursor = funcs.refresh_cursor(database=local_vars.maindb, connection=db_connection, existing_cursor=db_cursor, close_connection=True)
	t_layer = postprocess_zones(t_layer, db_cursor, args_native_qc, parent_layer)

	log.write("Running QC Historic Native Richness", 1)
	db_connection, db_cursor = funcs.refresh_cursor(database=local_vars.maindb, connection=db_connection, existing_cursor=db_cursor, close_connection=True)
	t_layer = postprocess_zones(t_layer, db_cursor, args_native_qc_historic, parent_layer)

	args_nonnative = [
		'SELECT Count(*) AS col_value FROM (select distinct Observations.Species_ID from Observations, Fish_Species where Observations.Zone_Id = ? And Observations.Species_ID = Fish_Species.FID and Fish_Species.Native = False and Observations.Presence_Type in (%s))' % (
		local_vars.current_obs_types), 'nonnative_richness', 'LONG']
	args_nonnative_qc = [
		'SELECT Count(*) AS col_value FROM (select distinct Observations.Species_ID from Observations, Fish_Species, Observation_Collections where Observations.Zone_Id = ? And Observations.Species_ID = Fish_Species.FID and Fish_Species.Native = False and Observations.Presence_Type in (%s) and Observations.OBJECTID = Observation_Collections.Observation_ID and Observation_Collections.Collection_ID in (%s))' % (
		local_vars.current_obs_types, local_vars.hq_collections), 'nonnative_qc_richness', 'LONG']

	if run_non_qc == "1" or run_non_qc == 1:
		log.write("Running NonNative Richness",1)
		db_connection, db_cursor = funcs.refresh_cursor(database=local_vars.maindb, connection=db_connection, existing_cursor=db_cursor, close_connection=True)
		t_layer = postprocess_zones(t_layer, db_cursor, args_nonnative, parent_layer)

	log.write("Running NonNative QC Richness",1)
	db_connection, db_cursor = funcs.refresh_cursor(database=local_vars.maindb, connection=db_connection, existing_cursor=db_cursor, close_connection=True)
	t_layer = postprocess_zones(t_layer, db_cursor, args_nonnative_qc, parent_layer)

	args_all = [
		'SELECT Count(*) AS col_value FROM (select distinct Observations.Species_ID from Observations, Fish_Species where Observations.Zone_Id = ? and Observations.Species_ID = Fish_Species.FID and Observations.Presence_Type in (%s))' % (
		local_vars.current_obs_types), 'all_richness', 'LONG']
	args_all_qc = [
		'SELECT Count(*) AS col_value FROM (select distinct Observations.Species_ID from Observations, Fish_Species, Observation_Collections where Observations.Zone_Id = ? and Observations.Species_ID = Fish_Species.FID and Observations.Presence_Type in (%s) and Observations.OBJECTID = Observation_Collections.Observation_ID and Observation_Collections.Collection_ID in (%s))' % (
		local_vars.current_obs_types, local_vars.hq_collections), 'all_qc_richness', 'LONG']

	if run_non_qc == "1" or run_non_qc == 1:
		log.write("Running Total Richness", 1)
		db_connection, db_cursor = funcs.refresh_cursor(database=local_vars.maindb, connection=db_connection, existing_cursor=db_cursor, close_connection=True)
		t_layer = postprocess_zones(t_layer, db_cursor, args_all, parent_layer)

	log.write("Running Total QC Richness", 1)
	db_connection, db_cursor = funcs.refresh_cursor(database=local_vars.maindb, connection=db_connection, existing_cursor=db_cursor, close_connection=True)
	t_layer = postprocess_zones(t_layer, db_cursor, args_all_qc, parent_layer)

	args_sensitive = [
		'SELECT Count(*) AS col_value FROM (select distinct Observations.Species_ID from Observations, Species_Aux, Fish_Species where Observations.Zone_Id = ?  and Observations.Presence_Type in (%s) And Observations.Species_ID = Fish_Species.FID and Observations.Species_ID = Species_Aux.FID and Species_Aux.Average < 3.01 and Species_Aux.Average > 0 and Fish_Species.Native = True)' % (
		local_vars.current_obs_types), 'sens_nat_rich', 'LONG']

	args_sensitive_qc = [
		'SELECT Count(*) AS col_value FROM (select distinct Observations.Species_ID from Observations, Species_Aux, Fish_Species, Observation_Collections where Observations.Zone_Id = ? And Observations.Presence_Type in (%s) And Observations.Species_ID = Fish_Species.FID and Observations.Species_ID = Species_Aux.FID and Species_Aux.Average < 3.01 and Species_Aux.Average > 0 and Fish_Species.Native = True and Observations.OBJECTID = Observation_Collections.Observation_ID and Observation_Collections.Collection_ID in (%s))' % (
		local_vars.current_obs_types, local_vars.hq_collections), 'sens_nat_qc_rich', 'LONG']

	if run_non_qc == "1" or run_non_qc == 1:
		log.write("Running Sensitivity Richness", 1)
		db_connection, db_cursor = funcs.refresh_cursor(database=local_vars.maindb, connection=db_connection, existing_cursor=db_cursor, close_connection=True)
		t_layer = postprocess_zones(t_layer, db_cursor, args_sensitive, parent_layer)

	log.write("Running QC Sensitivity Richness", 1)
	db_connection, db_cursor = funcs.refresh_cursor(database=local_vars.maindb, connection=db_connection, existing_cursor=db_cursor, close_connection=True)
	t_layer = postprocess_zones(t_layer, db_cursor, args_sensitive_qc, parent_layer)

	log.write("Getting Sensitive Species Assemblages", 1)
	args_sensitive_assem = [function_arg('_subfunction_convert_species', (
	'select distinct Observations.Species_ID as col_value from Observations, Species_Aux, Fish_Species where Observations.Zone_Id = ? and Observations.Presence_Type in (%s) And Observations.Species_ID = Fish_Species.FID and Observations.Species_ID = Species_Aux.FID and Species_Aux.Average < 3.01 and Species_Aux.Average > 0 and Fish_Species.Native = True' % (
	local_vars.current_obs_types),)), 'sens_nat_assem', 'TEXT']
	args_sensitive_qc_assem = [function_arg('_subfunction_convert_species', (
	'select distinct Observations.Species_ID as col_value from Observations, Species_Aux, Fish_Species, Observation_Collections where Observations.Zone_Id = ? and Observations.Presence_Type in (%s) And Observations.Species_ID = Fish_Species.FID and Observations.Species_ID = Species_Aux.FID and Species_Aux.Average < 3.01 and Species_Aux.Average > 0 and Fish_Species.Native = True and Observations.OBJECTID = Observation_Collections.Observation_ID and Observation_Collections.Collection_ID in (%s)' % (
	local_vars.current_obs_types, local_vars.hq_collections),)), 'sens_nat_qc_assem', 'TEXT']
	t_layer = postprocess_zones(t_layer, db_cursor, args_sensitive_assem, parent_layer)
	t_layer = postprocess_zones(t_layer, db_cursor, args_sensitive_qc_assem, parent_layer)

	log.write("Generating Sensitivity Statistics", 1)
	db_connection, db_cursor = funcs.refresh_cursor(database=local_vars.maindb, connection=db_connection, existing_cursor=db_cursor, close_connection=True)
	t_layer = sensitivity_stats(t_layer, db_cursor, (1,), parent_layer)  # run it the first time with qc flag

	if run_non_qc == "1" or run_non_qc == 1:
		db_connection, db_cursor = funcs.refresh_cursor(database=local_vars.maindb, connection=db_connection, existing_cursor=db_cursor, close_connection=True)
		t_layer = sensitivity_stats(t_layer, db_cursor, (0,), parent_layer)  # second time without

	log.write("Generating Richness Differences", 1)
	db_connection, db_cursor = funcs.refresh_cursor(database=local_vars.maindb, connection=db_connection, existing_cursor=db_cursor, close_connection=True)
	t_layer = richness_difference(t_layer, db_cursor, ("Native_Fish", 5), parent_layer)

	# beta diversity measures
	log.write("Generating Beta Diversities", 1)
	args_beta_qc_native = ['native', 'nodams', 'diversity_jaccard', 'qc']
	args_beta_native = ['native', 'nodams', 'diversity_jaccard', 'no']
	args_beta_qc_all = ['nonnative', 'nodams', 'diversity_jaccard', 'qc']
	args_beta_all = ['nonnative', 'nodams', 'diversity_jaccard', 'no']

	db_connection, db_cursor = funcs.refresh_cursor(database=local_vars.maindb, connection=db_connection, existing_cursor=db_cursor, close_connection=True)
	t_layer = diversity(t_layer, db_cursor, args_beta_qc_native, parent_layer)

	if run_non_qc == "1" or run_non_qc == 1:
		db_connection, db_cursor = funcs.refresh_cursor(database=local_vars.maindb, connection=db_connection, existing_cursor=db_cursor, close_connection=True)
		t_layer = diversity(t_layer, db_cursor, args_beta_native, parent_layer)
	db_connection, db_cursor = funcs.refresh_cursor(database=local_vars.maindb, connection=db_connection, existing_cursor=db_cursor, close_connection=True)
	t_layer = diversity(t_layer, db_cursor, args_beta_qc_all, parent_layer)

	if run_non_qc == "1" or run_non_qc == 1:
		db_connection, db_cursor = funcs.refresh_cursor(database=local_vars.maindb, connection=db_connection, existing_cursor=db_cursor, close_connection=True)
		t_layer = diversity(t_layer, db_cursor, args_beta_all, parent_layer)

	log.write("Getting Downstream Diversities", 1)
	db_connection, db_cursor = funcs.refresh_cursor(database=local_vars.maindb, connection=db_connection, existing_cursor=db_cursor, close_connection=True)
	t_layer = get_downstream_diversities(t_layer, db_cursor, None, parent_layer)

	return t_layer


def sensitivity_stats(zones_layer, db_cursor, args, parent_layer):
	"""

	:param zones_layer:
	:param db_cursor:
	:param args:
	:param parent_layer:
	:return:
	"""

	stats_query = query()

	if args and (args[0] == "1" or args[0] == 1):
		qc_flag = True
		qc_insert = "qc_"
		log.write("QC = True", 1)
	else:
		qc_flag = False
		qc_insert = ""
		log.write("QC = False", 1)

	def make_query(qc_f):
		stats_query = query()
		stats_query.set_defaults(zonal_callback=True, qc=qc_f)
		stats_query.tables.append("Species_Aux")
		stats_query.where_clauses.append("Species_Aux.FID = Species.FID")

		stats_query.base = 'SELECT AVG(Species_Aux.Average) AS col_value from (select distinct Species_Aux.FID, Species_Aux.Average'
		stats_query.suffix = ")"

		return stats_query

	stats_query = make_query(qc_flag)
	stats_args = [stats_query.compose(), 'avg_%ssensitivity_score' % qc_insert, 'DOUBLE']
	log.write("sensitivity query is: %s" % stats_query.compose(), 1)
	t_layer = postprocess_zones(zones_layer, db_cursor, stats_args, parent_layer)

	stats_query2 = make_query(qc_flag)
	stats_query2.where_clauses.append("Species_Aux.Average <= 3 and Species_Aux.Average >= 0")  # change the last item
	stats_args2 = [stats_query2.compose(), 'avg_%ssensitivity_under3' % qc_insert, 'DOUBLE']
	log.write("sensitivity query is: %s" % stats_query2.compose(), 1)
	t_layer = postprocess_zones(t_layer, db_cursor, stats_args2, parent_layer)

	stats_query3 = make_query(qc_flag)
	stats_query3.where_clauses.append("Species_Aux.Average <= 2 and Species_Aux.Average >= 0")
	stats_args3 = [stats_query3.compose(), 'avg_%ssensitivity_under2' % qc_insert, 'DOUBLE']
	log.write("sensitivity query is: %s" % stats_query3.compose(), 1)
	t_layer = postprocess_zones(t_layer, db_cursor, stats_args3, parent_layer)

	stats_query4 = make_query(qc_flag)
	stats_query4.where_clauses.append("Species_Aux.Average <= 1 and Species_Aux.Average >= 0")
	stats_args4 = [stats_query4.compose(), 'avg_%ssensitivity_under1' % qc_insert, 'DOUBLE']
	log.write("sensitivity query is: %s" % stats_query4.compose(), 1)
	t_layer = postprocess_zones(t_layer, db_cursor, stats_args4, parent_layer)

	return t_layer


def get_downstream_diversities(t_layer, db_cursor, cb_args, parent_layer):
	"""
		Standard callback - takes a layer of hucs, gets the assemblage for all hucs, then adds two fields with that data
		1) contains the downstream assemblage and
		2) contains the count of downstream taxa

	@param t_layer:
	@param db_cursor:
	@param cb_args:
	@param parent_layer:
	@return: t_layer
	"""
	global diversity_network
	if not diversity_network:
		diversity_network = get_diversity_into_network(False, db_cursor, True, True, "simple", t_layer)

	log.write("Getting Downstream Diversities", 1)

	args = ("downstream", "",)
	func_arg = function_arg("_get_assemblage_for_row", args)
	postprocess_zones(t_layer, db_cursor, (func_arg, "downstream_assemblage", "TEXT"), parent_layer)
	args2 = ("downstream", "count",)
	func_arg2 = function_arg("_get_assemblage_for_row", args2)
	postprocess_zones(t_layer, db_cursor, (func_arg2, "downstream_count", "DOUBLE"), parent_layer)

	return t_layer


def _get_assemblage_for_row(t_layer, db_cursor, args, parent_layer):
	"""
		Given a huc in args[1], it returns the assemblage of the downstream huc

	@param t_layer:
	@param db_cursor:
	@param args:
		A three item tuple:
			0: location to get assemblage for - default is the current zone. If args[0] == "downstream" then the downstream
				zone assemblage is returned
			1: return assemblage or count of assemblage? returns assemblage by default and count when args[1] == "count"
			2: whether to return common name or FID values for assemblage. returns common name by default and FIDs
				when args[2] == "FID"

	@param parent_layer:
	@return:
	"""
	#TODO: Document - this is a kluge - Nick was sick and wrote it in a rush

	huc_id = args[1]
	func_args = args[0].argument

	if func_args and len(func_args) > 1 and func_args[1] == "count":
		# set this up once so it's not done multiple times below
		empty = 0
	else:
		empty = []

	try:
		if func_args and len(func_args) > 0 and func_args[0] == "downstream":
			downstream_id = diversity_network[huc_id].downstream
			if downstream_id not in diversity_network: # doesn't exist - return 0 or empty list
				return empty
			#check if it exists, and return an empty list if it doesn't to signify unknown or none - this way, we don't run up into a "NoneType has no attribute __iter__" later
			if func_args and len(func_args) > 1 and func_args[1] == "count":
				return len(diversity_network[downstream_id].assemblage)
			else:
				return_data = diversity_network[downstream_id].assemblage
		else:
			if huc_id not in diversity_network:
				return empty # check if it exists, and return an empty list if it doesn't to signify unknown or none - this way, we don't run up into a "NoneType has no attribute __iter__" later
			if func_args and len(func_args) > 1 and func_args[1] == "count":
				return len(diversity_network[huc_id].assemblage)
			else:
				return_data = diversity_network[huc_id].assemblage
	except IndexError:
		# not sure why we're doing this on error handling
		return diversity_network[huc_id].assemblage

	if func_args and len(func_args) > 2 and func_args[2] == "FID":
		return return_data
	else:
		return diversity_assemblage_to_cname(return_data)


def write_results_by_key(layer, results, layer_field, results_field, layer_key="HUC_12", results_key = "HUC_12",):

	arc_curs = arcpy.UpdateCursor(layer)

	# make a lookup structure for the data
	results_dict = {}
	for row in results:
		results_dict[row.__getattribute__(results_key)] = row.__getattribute__(results_field)

	# for every row
	for row in arc_curs:
		cur_key = row.getValue(layer_key)

		if not cur_key in results_dict:  # skip it if it's not there
			continue

		row.setValue(layer_field, results_dict[cur_key])
		arc_curs.updateRow(row)


def split_postprocess_cbargs(cb_args, layer):
		# structure the queries out of the arguments
	t_count = 0
	l_queries = []
	while t_count < len(cb_args):
		t_query = empty_object()
		t_query.query = cb_args[t_count]
		t_query.column = cb_args[t_count + 1]
		t_query.col_type = cb_args[t_count + 2]

		if isinstance(t_query.query, function_arg): # if it's a class instance, then it's a function_arg
			t_query.run_function = True
		else:
			t_query.run_function = False

		try:
			if string.upper(t_query.col_type) == "TEXT":
				# we want to set the length on text fields
				arcpy.AddField_management(layer, t_query.column, t_query.col_type, "", "", 10000) # add the column
			else:
				arcpy.AddField_management(layer, t_query.column, t_query.col_type)
		except:
			raise local_vars.MappingError(
				"Unable to run callback postprocess_zones - error adding field %s" % t_query.column)

		# replace entities in the query
		#t_query.query = replace_variables(t_query.query,parent_layer)

		l_queries.append(t_query)

		t_count += 3  # skip the next 2 records

	return l_queries


def coerce_value_from_arc_type(val, col_type):
	if val is None:  # if it's None, then we don't care what type it is, just set it to None so arcpy can make it NULL
		row_val = None
	elif string.upper(col_type) == "TEXT":  # coerce types
		if isiterable(val):
			row_val = stringify(val)  # stringify is a nicer way of going about it for iterables
		else:
			row_val = str(val)
	elif string.upper(col_type) == "LONG":
		row_val = int(val)
	elif string.upper(col_type) == "FLOAT" or string.upper(col_type) == "DOUBLE":
		row_val = float(val)
	else:
		row_val = val

	return row_val


def add_field(zones_layer, db_cursor, cb_args, parent_layer):
	"""
		Very similar to postprocess_zones except it executes a query once, and joins a result to the output layer
		Meant to be used for summary queries that summarize by HUC12

	:param zones_layer:
	:param db_cursor:
	:param cb_args:
	:param parent_layer:
	:return:
	"""


	# TODO: Warning, this code will leave this dataset in memory - a fixed number of these can be called in any given session
	new_name = generate_fast_filename("callback_add_field", return_full=True)
	arcpy.CopyFeatures_management(zones_layer, new_name)

	l_queries = split_postprocess_cbargs(cb_args, new_name)  # this also adds the fields to the layers

	for l_query in l_queries:  # for every input query_object, runs the query, and passes results to function to be written out
		results = db_cursor.execute(l_query.query)
		write_results_by_key(new_name, results, l_query.column, "col_value", results_key="Zone_ID")

	return new_name



def postprocess_zones(zones_layer, db_cursor, cb_args, parent_layer):
	'''
		Takes each record in the zones_layer and runs queries related to them.
		Queries should specify the place to include the zone with sql bind variables "?" and the variable operators specified by object location relative to parent_layer (eg: {custom_query.bind_var}) are optional
		Flexible:{object path from parent layer being worked on} - example {custom_query.bind_var} would return parent_layer.custom_query.bind_var and dump it into the query at that location
		Argument Format: query,column_name,column_data_type,query,column_name,column_data_type,query...etc
		Notes: the value to be placed in the column for a given query should be selected "AS col_value" - that's what the following code will be looking to retrieve
	'''

	arcpy.env.overwriteOutput = True

	log.write("Postprocessing zones - this may take some time!", 1)
	#log.write("%s queries for layer" % len(cb_args) / 2)

	l_name = arcpy.CreateUniqueName('postprocessing_temp', local_vars.workspace)
	l_temp_file = os.path.join(local_vars.workspace, l_name)
	arcpy.CopyFeatures_management(zones_layer, l_temp_file) #copy any selected zones out

	l_queries = split_postprocess_cbargs(cb_args, l_temp_file)

	rows = arcpy.UpdateCursor(l_temp_file)

	# for each row
	for row in rows:

		#check that it's a HUC
		if row.getValue(local_vars.huc_field) is None:
			print "skipping row..."
			continue

		# then run the queries
		for query in l_queries:
			if query.run_function:
				# if we're supposed to run a function instead of just executing the queries
				val = query.query.function_obj(zones_layer, db_cursor,
				                               [query.query, row.getValue(local_vars.huc_field), cb_args], parent_layer)
			else:
				db_cursor.execute(query.query, row.getValue(local_vars.huc_field))
				l_result = db_cursor.fetchone()  # just get me the first row. There only should be one anyway...
				val = l_result.col_value

			row_val = coerce_value_from_arc_type(val, query.col_type)
			row.setValue(query.column, row_val)

		rows.updateRow(row)  # save it!

	del row
	del rows  # cleanup
	arcpy.env.overwriteOutput = False

	arcpy.Delete_management(zones_layer)  # kill the previous version
	arcpy.MakeFeatureLayer_management(l_temp_file, zones_layer)  # and read in the new one

	# return it
	return zones_layer


def stringify(in_list, delimiter=", "):
	"""
	Turns a list into a string without the "u" or brackets
	@param in_list: list to process
	@param delimiter: string denoting what to use to separate items
	@return: string
	"""
	l_string = ""

	for item in in_list:
		l_string += "%s%s" % (str(item), delimiter)

	return l_string


def _subfunction_convert_species(zones_layer, db_cursor, args, parent_layer):
	'''an assemblage based function. If you pass in a query to the function_arg that returns an assemblage in pisces codes, this will convert them to common names first'''
	#args[0] contains the function_arg object
	#args[1] contains the bind variable (the huc)
	#args[2] contains all of the args to postprocess_zones

	query = args[0].argument[0]

	results = db_cursor.execute(query, args[1])
	species = []
	for row in results:
		species.append(row.col_value)
	return str(diversity_assemblage_to_cname(species))


def replace_variables(replacement_string, parent_layer):
	while True: # find any object requests in the query - this is bad way to do this, but it "should" break on the first non_match

		if type(
			replacement_string) != "str": # if we pass in an object instead of a string, don't search it - it's their own damn fault for being complicated. They can search themselves
			break

		m_object = re.search("({.+?})", replacement_string) # {.+?}

		if m_object == None: # if we don't have any replacement variables, break
			break
		match_item = m_object.group(0)

		#get the path to the object
		if "." in match_item: # if it has multiple parts, split it into a tuple
			object_parts = match_item.split(".")
		else: # otherwise, make our own one item tuple
			object_parts = (match_item,) # make it a single item tuple

		replace_object = parent_layer # start with the layer object
		for part in object_parts: # then for each portion of the object path
			try:
				replace_object = replace_object.__dict__[
					part] # get that part and set it as the object to replace - effectively, move "down" the object
			except: # I see this being very possibly called - specifying a bad path would cause this (IndexError?)
				raise local_vars.MappingError(
					"Unable to run callback postprocess_zones - couldn't get custom entity specified in {%s}" % match_item)

		replacement_string.replace("{%s}" % match_item, replace_object)

	return replacement_string


def richness(zones_layer, db_cursor, cb_args, parent_layer): #layer callback
	"""
		An old richness function - superceded by passing the correct parameters to postprocess_zones
	@param zones_layer:
	@param db_cursor:
	@param cb_args:
	@param parent_layer:
	@return:
	"""

	arcpy.env.overwriteOutput = True

	log.write("Building species richness layer - this may take some time!", 1)

	l_temp_file = os.path.join(local_vars.workspace, "richness")
	arcpy.CopyFeatures_management(zones_layer, l_temp_file) #copy any selected zones out

	arcpy.AddField_management(l_temp_file, "Richness", "LONG") # add the column

	rows = arcpy.UpdateCursor(l_temp_file)

	if not (cb_args == None or cb_args is False):
		sql_ext = cb_args
	else:
		sql_ext = ""

	log.write("sql extension on callback: %s" % sql_ext, 1)

	for row in rows:
		if row.HUC_12 == None:
			print "skipping row..."
			continue

		l_sql = "SELECT Count(*) AS l_count FROM (select distinct Observations.Species_ID from Observations, Fish_Species where Observations.Zone_Id = ? and Observations.Species_ID = Fish_Species.FID " + sql_ext + ")"
		db_cursor.execute(l_sql, row.HUC_12)

		l_result = db_cursor.fetchone() # just get me the first row. There only should be one anyway...
		row.Richness = l_result.l_count # the first row of the result and the first item in that row

		rows.updateRow(row) # save it!

	del row
	del rows # cleanup
	arcpy.env.overwriteOutput = False

	# read it back in
	zone_layer = "Richness_Zones"
	arcpy.MakeFeatureLayer_management(l_temp_file, zone_layer)

	# return it
	return zone_layer


def _compose_simple_query(sql_base, sql_tables, where_clauses):
	sql_tables_str = ""
	for table in sql_tables:
		sql_tables_str += " %s," % table
	sql_tables_str = sql_tables_str[:-1] # chop off the last ,

	sql_where_str = " where"
	for where in where_clauses:
		sql_where_str += " %s and" % where
	sql_where_str = sql_where_str[:-3] # chop off the last "and"

	return sql_base + sql_tables_str + sql_where_str


def richness_difference(zones_layer, db_cursor, args, parent_layer):
	species_lim = get_arg(args, 0, "Native_Fish")
	species_limit = funcs.text_to_species_list(species_lim)  # pass it into the species/group parser
	observations_limit = get_arg(args, 1,
	                             local_vars.hq_collections)  # tuple of one and three represents the defaultobservation sets we're interested in
	obs_cur = get_arg(args, 2, local_vars.current_obs_types)
	obs_hist = get_arg(args, 3, local_vars.historic_obs_types)

	sql_tables = ["Observations"]
	where_clauses = ["Observations.Presence_Type in (%s)" % obs_cur, "Observations.Zone_ID = ?"]

	if species_limit:
		spp_string = ""
		for item in species_limit:
			spp_string += "'%s'," % item
		spp_string = spp_string[:-1]
		where_clauses.append("Species_ID in (%s)" % spp_string)

	if observations_limit:
		sql_tables.append("Observation_Collections")
		where_clauses += ["Observations.OBJECTID = Observation_Collections.Observation_ID",
		                  "Observation_Collections.Collection_ID in (%s)" % (
		                  str(observations_limit))]  # chop the list brackets off the list
	else:
		log.write(
			"No species selected for limiting - if you intended to limit to a group or list of species, check your callback arguments",
			True)

	run_losses = get_arg(args, 4, True)

	current_sql_query = _compose_simple_query(
		"select count(*) as col_value from (select distinct Observations.Species_ID from ", sql_tables,
		where_clauses) + ")"  # close the subquery
	loss_gain_current = _compose_simple_query("select distinct Observations.Species_ID from ", sql_tables,
	                                          where_clauses)
	where_clauses[0] = "Observations.Presence_Type in (%s)" % obs_hist  # replace the first portion of the clause
	historic_sql_query = _compose_simple_query(
		"select count(*) as col_value from (select distinct Observations.Species_ID from ", sql_tables,
		where_clauses) + ")"  # close the subquery
	loss_gain_historic = _compose_simple_query("select distinct Observations.Species_ID from ", sql_tables,
	                                           where_clauses)

	log.write("Current richness query is %s" % current_sql_query)
	log.write("Historic richness query is %s" % historic_sql_query)
	log.write("Losses/Gains historic query (partial - it's passed through another function) is %s" % loss_gain_historic)
	log.write("Losses/Gains current query (partial - it's passed through another function) is %s" % loss_gain_current)

	current_field = "current_richness"
	historic_field = "historic_richness"
	current_assemblage_field = "current_assemblage"
	historic_assemblage_field = "historic_assemblage"
	current_args = [current_sql_query, current_field, "LONG"]
	hist_args = [historic_sql_query, historic_field, "LONG"]
	hist_assemblage_args = [loss_gain_historic, historic_assemblage_field, "TEXT"]

	losses_field = "losses"
	gains_field = "gains"
	losses_func_arg = function_arg("subprocess_losses", (loss_gain_historic, loss_gain_current, 0, False))
	losses_list_arg = function_arg("subprocess_losses", (loss_gain_historic, loss_gain_current, 0, True))
	gains_func_arg = function_arg("subprocess_losses", (loss_gain_historic, loss_gain_current, 1, False))
	gains_list_arg = function_arg("subprocess_losses", (loss_gain_historic, loss_gain_current, 1, True))
	current_assemblage_arg = function_arg("subprocess_assemblage", (loss_gain_current,))
	historic_assemblage_arg = function_arg("subprocess_assemblage", (loss_gain_historic,))

	current_assemblage_args = [current_assemblage_arg, current_assemblage_field, "TEXT"]
	hist_assemblage_args = [historic_assemblage_arg, historic_assemblage_field, "TEXT"]

	losses_args = [losses_func_arg, losses_field, "LONG"]
	gains_args = [gains_func_arg, gains_field, "LONG"]

	losses_list_args = [losses_list_arg, losses_field + "_list", "TEXT"]
	gains_list_args = [gains_list_arg, gains_field + "_list", "TEXT"]

	log.write("Running current richness and assemblage", True)
	t_layer = postprocess_zones(zones_layer, db_cursor, current_args, parent_layer)
	t_layer = postprocess_zones(t_layer, db_cursor, current_assemblage_args, parent_layer)
	log.write("Running historic richness and assemblage", True)
	t_layer = postprocess_zones(t_layer, db_cursor, hist_args, parent_layer)
	t_layer = postprocess_zones(t_layer, db_cursor, hist_assemblage_args, parent_layer)

	if run_losses:
		log.write("Running losses", True)
		t_layer = postprocess_zones(t_layer, db_cursor, losses_args, parent_layer)
		log.write("Running losses as list output", True)
		t_layer = postprocess_zones(t_layer, db_cursor, losses_list_args, parent_layer)
		log.write("Running gains", True)
		t_layer = postprocess_zones(t_layer, db_cursor, gains_args, parent_layer)
		log.write("Running gains as list output", True)
		t_layer = postprocess_zones(t_layer, db_cursor, gains_list_args, parent_layer)

	field_name = "richness_difference"
	arcpy.AddField_management(t_layer, field_name, "LONG")

	t_curs = arcpy.UpdateCursor(t_layer)
	for row in t_curs:
		try:
			val = int(row.getValue(current_field) - row.getValue(historic_field))
		except:
			val = -9999 # a value that will be noticed

		try:
			row.setValue(field_name, val)
			t_curs.updateRow(row)
		except:
			log.error("Couldn't update a row with the richness difference")

	del t_curs
	return t_layer


def subprocess_losses(zones_layer, db_cursor, cb_args, parent_layer):
	"""
	a subfunction for use with postprocess_zones.
	This entire function could be done in a single query, but Access won't run it. So...here we are.
	#cb_args[0] contains the function_arg object
	#cb_args[1] contains the bind variable (the huc)
	#cb_args[2] contains all of the args to postprocess_zones
	"""

	query_hist = cb_args[0].argument[0] # the first item in args is a tuple, the second is the args originally passed to postprocess_zones
	query_current = cb_args[0].argument[1]
	order = cb_args[0].argument[2]
	in_list_form = cb_args[0].argument[3]
	if in_list_form is None:
		in_list_form = False
	query_bind = cb_args[1]

	if order not in (0, 1):  # order specifies whether we mean losses or gains
		log.error(
			"Order parameter not correctly specified. It must be defined and it must be either 0 (historic) or 1 (current)")
		raise ValueError(
			"Order parameter not correctly specified. It must be defined and it must be either 0 (historic) or 1 (current)")

	results = db_cursor.execute(query_hist, query_bind)
	list_hist = []

	for item in results:
		list_hist.append(item.Species_ID)

	results_cur = db_cursor.execute(query_current, query_bind)
	list_cur = []

	for item in results_cur:
		list_cur.append(item.Species_ID)

	if order == 1:  # gains
		new_set = frozenset(list_cur) - frozenset(list_hist) # set of species only in current
	else: # losses = 0 - can safely be written as else due to error checking up higher - faster this way too and satisfies PyCharm's error checker
		new_set = frozenset(list_hist) - frozenset(list_cur) # set of species only in historic

	distinct_items = list(new_set)

	if in_list_form:
		common_names = []
		for item in distinct_items:
			common_names.append(local_vars.all_fish[item].species)
		return common_names
	else:
		return len(distinct_items)


def subprocess_assemblage(zones_layer, db_cursor, cb_args, parent_layer):
	'''a subfunction for use with postprocess_zones.
	This entire function could be done in a single query, but Access won't run it. So...here we are.'''
	#cb_args[0] contains the function_arg object
	#cb_args[1] contains the bind variable (the huc)
	#cb_args[2] contains all of the args to postprocess_zones

	query = cb_args[0].argument[
		0] # the first item in args is a tuple, the second is the args originally passed to postprocess_zones
	query_bind = cb_args[1]
	try:
		in_list_form = cb_args[0].argument[1]
	except:
		in_list_form = True
	if in_list_form is None:
		in_list_form = True

	results = db_cursor.execute(query, query_bind)
	distinct_items = []

	for item in results:
		distinct_items.append(item.Species_ID)

	if in_list_form:
		common_names = []
		for item in distinct_items:
			common_names.append(local_vars.all_fish[item].species)
		return common_names
	else:
		return len(distinct_items)


def genus_family_richness(zones_layer, db_cursor, gf, parent_layer): #layer callback
	"""

	@param zones_layer:
	@param db_cursor:
	@param gf:
	@param parent_layer:
	@return:
	"""
	arcpy.env.overwriteOutput = True

	log.write("Building family richness layer - this may take some time!", 1)

	l_temp_file = os.path.join(local_vars.workspace, "richness")
	arcpy.CopyFeatures_management(zones_layer, l_temp_file) #copy any selected zones out

	arcpy.AddField_management(l_temp_file, "Richness", "LONG") # add the column

	rows = arcpy.UpdateCursor(l_temp_file)

	for row in rows:
		l_sql = "SELECT Count(*) AS l_count FROM (select distinct Family from Fish_Species where FID = (select distinct Species_ID from Observations where Zone_Id = ?))"
		db_cursor.execute(l_sql, row.HUC_12)

		l_result = db_cursor.fetchone() # just get me the first row. There only should be one anyway...
		row.Richness = l_result.l_count # the first row of the result and the first item in that row

		rows.updateRow(row) # save it!

	del row
	del rows # cleanup
	arcpy.env.overwriteOutput = False

	# read it back in
	zone_layer = "Richness_Zones"
	arcpy.MakeFeatureLayer_management(l_temp_file, zone_layer)

	# return it
	return zone_layer


def make_tooltip_column(zones_layer, db_cursor, args):
	arcpy.env.overwriteOutput = True

	log.write("Building tooltip", 1)

	l_temp_file = os.path.join(local_vars.workspace, "tooltip")
	arcpy.CopyFeatures_management(zones_layer, l_temp_file) #copy any selected zones out

	arcpy.AddField_management(l_temp_file, "Tooltip", "LONG") # add the column

	rows = arcpy.UpdateCursor(l_temp_file)

	for row in rows:
		if row.HUC_12 == None:
			print "skipping row..."
			continue

		l_sql = "SELECT DISTINCT Observation_Sets.* FROM Observation_Sets, Observations WHERE Observations.Species_ID = ? and Observations.Zone_ID = ? and Observations.Set_ID = Observation_Sets.Set_ID;"
		db_cursor.execute(l_sql,
		                  row.HUC_12) # TODO: Needs to use a query brought in by argument and needs a way to bring in the species and pass it as a bind variable - need a robust system for operators in the text that get replaced before the bind arguments are passed in.

		l_result = db_cursor.fetchone() # just get me the first row. There only should be one anyway...
		row.Richness = l_result.l_count # the first row of the result and the first item in that row

		rows.updateRow(row) # save it!

	del row
	del rows  # cleanup
	arcpy.env.overwriteOutput = False

	# read it back in
	zone_layer = "Richness_Zones"
	arcpy.MakeFeatureLayer_management(l_temp_file, zone_layer)

	# return it
	return zone_layer


def representation(zones_layer, db_cursor, cb_args, parent_layer):
	'''just an alias for clipped_layer_overlay'''
	return clipped_layer_overlay(zones_layer, db_cursor, cb_args, parent_layer)


def clipped_layer_overlay(zones_layer, db_cursor, cb_args, parent_layer):
	log.write("Creating Layer Representation", 1)
	if not cb_args or len(cb_args) == 0: # if we don't have an argument, default to all_rivers
		clip_layer = os.path.join(local_vars.geo_aux, "all_rivers")
	else: # callback args come in as a list, so pull the first item
		clip_layer = cb_args[0]

	clipped_layer = "clipped_layer"
	if arcpy.Exists(clipped_layer):
		# if it's already in memory, then destroy it
		arcpy.Delete_management(clipped_layer)

	log.write("Running callback clipped_layer_overlay for feature %s" % clip_layer)

	#unique_arc_name = arcpy.CreateUniqueName("temp_",local_vars.workspace) # generate a unique name for the zones dissolve
	#arcpy.Dissolve_management(zones_layer,unique_arc_name) # dissolve the HUCs so that the input is

	unique_arc_name_2 = arcpy.CreateUniqueName("temp_mapping_result_", local_vars.workspace)
	arcpy.Clip_analysis(clip_layer, zones_layer,
	                    unique_arc_name_2) # clip the input to the dissolved zones and put it in unique_arc_name_2

	t_layer = arcpy.MakeFeatureLayer_management(unique_arc_name_2, clipped_layer)

	# return clip_layer
	return clipped_layer


def connectivity(zones_layer, db_cursor, args, map_layer):
	att_types = args[0]
	attributes = args[1:]

	arcpy.env.overwriteOutput = True

	new_layer = os.path.join(local_vars.workspace, "connectivity_temp")

	arcpy.CopyFeatures_management(zones_layer, new_layer) # copy it out
	arcpy.MakeFeatureLayer_management(new_layer,
	                                  zones_layer) # copy it back - need to do this to not modify original hucs

	arcpy.AddField_management(zones_layer, "connectivity_upstream_count", "LONG")

	select_vars = ""
	for att in attributes:
		# for every attribute, add a column to the layer and add it to the list of variables we want to collect info on in order to select appropriately
		try:
			arcpy.AddField_management(zones_layer, "connectivity_%s" % att, att_types)
		except:
			raise local_vars.MappingError(
				"Couldn't add field in callback 'connectivity' - make sure you are specifying an appropriate data type for the column")

		select_vars += "%s; " % att

	huc_cursor = arcpy.UpdateCursor(zones_layer)
	sys.setrecursionlimit(6000) # set it to far higher than the number of HUCs - we shouldn't even get close to it...

	h_num = 0
	for huc in huc_cursor:
		h_num += 1
		print "%s - Next" % h_num
		connectivity_recurse(zones_layer, huc, select_vars, attributes, huc_cursor)
		del huc

	# this function short circuits if we've already evaluated a given huc, so the attributes and upstream of any given huc will only be run once for speed

	del huc_cursor

	log.write("Deleting EVERYTHING from table connectivity to insert new values!", 1)
	db_cursor.execute("delete * from Connectivity")

	global connectivity_index
	sql = "insert into Connectivity (ZoneD,ZoneU) values (?,?)"
	for huc in connectivity_index.keys():
		for up_huc in connectivity_index[huc]:
			db_cursor.execute(sql, huc, up_huc)

	return zones_layer


def connectivity_recurse(zones_layer, huc, select_vars, atts, huc_cursor): # atts = attributes to look at

	global connectivity_index

	if huc.getValue(
		"connectivity_upstream_count") is not None: # if attributes are already set then we don't need to continue upstream of it - just use the current value
		return_obj = empty_object()
		return_obj.connectivity_upstream_count = huc.getValue("connectivity_upstream_count")
		for item in atts: # it's an object just for ease of access
			return_obj.__dict__[item] = huc.getValue(item) # add the value for the current huc
			return_obj.__dict__["connectivity_%s" % item] = huc.getValue(
				"connectivity_%s" % item) # and the value for all upstream hucs
			return_obj.hucs = connectivity_index[huc.HUC_12] # set the list of upstream hucs
		return return_obj

	upstream_hucs = arcpy.UpdateCursor(zones_layer, "\"HU_12_DS\" = '%s'" % huc.getValue("HUC_12"), "", "",
	                                   "") # needs to be an update cursor because we need to use the cursor that a row came from to update
	#select huc_12, attributes from local_vars.HUCS where HUC12FullState.HU_12_DS = huc_id

	conn_count = 0 # the number of hucs upstream
	return_obj = empty_object()
	return_obj.hucs = []

	for item in atts:
		return_obj.__dict__[item] = huc.getValue(item) # add the value for the current huc
		return_obj.__dict__["connectivity_%s" % item] = None

	hucs_upstream = False
	# we have hucs upstream!
	for up_huc in upstream_hucs:
		hucs_upstream = True
		conn_count += 1 # each huc here is an additional huc

		# Recurse!
		upstream_atts = connectivity_recurse(zones_layer, up_huc, select_vars, atts, upstream_hucs)

		conn_count += upstream_atts.connectivity_upstream_count # now also add the number of upstream hucs the upstream huc hadd
		for item in atts:
			if upstream_atts.__dict__["connectivity_%s" % item] is None:
				upstream_atts.__dict__["connectivity_%s" % item] = 0
			if upstream_atts.__dict__[item] is None:
				upstream_atts.__dict__[item] = 0
			if return_obj.__dict__["connectivity_%s" % item] is None:
				return_obj.__dict__["connectivity_%s" % item] = 0

			return_obj.__dict__["connectivity_%s" % item] += upstream_atts.__dict__["connectivity_%s" % item] + upstream_atts.__dict__[item] # add the value of the upstream huc to its upstream values and assign it to the spot on this huc
			return_obj.hucs.append(up_huc.HUC_12)
			return_obj.hucs.extend(upstream_atts.hucs)

	if hucs_upstream is False:  # if nothing is upstream, then we didn't enter the loop!
		for item in atts:
			return_obj.__dict__[item] = huc.getValue(item) # add the value for the current huc
			return_obj.__dict__[
				"connectivity_%s" % item] = 0 # WARNING - this doesn't scale to different data types...

	return_obj.connectivity_upstream_count = conn_count
	huc.setValue("connectivity_upstream_count", conn_count)
	for field in atts: #set the fields to the row
		try:
			huc.setValue("connectivity_%s" % field, return_obj.__dict__["connectivity_%s" % field])
		except:
			log.error("Unable to set value of field connectivity_%s for huc %s" % (field, huc.HUC_12))

		try:
			connectivity_index[huc.HUC_12] = return_obj.hucs
		except:
			log.error("Unable to save upstream hucs")

	huc_cursor.updateRow(huc)
	return return_obj


def network_distance(zones_layer, db_cursor, args, map_layer):
	log.write("Getting network distances", 1)

	# set up the lookup
	zones_info = {}
	zones_cursor = arcpy.SearchCursor(zones_layer)
	for zone in zones_cursor:
		zones_info[zone.HUC_12] = zone.HU_12_DS
	del zones_cursor

	if args[0] is None:
		return zones_layer

	arcpy.env.overwriteOutput = True

	new_layer = os.path.join(local_vars.workspace, "network_dist_temp")

	arcpy.CopyFeatures_management(zones_layer, new_layer) # copy it out
	arcpy.MakeFeatureLayer_management(new_layer,
	                                  zones_layer) # copy it back - need to do this to not modify original hucs
	arcpy.AddField_management(zones_layer, "network_distance", "INTEGER")

	if not type(
		args) == 'list': # if we don't have a second huc, then just get the distance from every huc in the layer to this one
		l_curs = arcpy.UpdateCursor(zones_layer)
		ind = 0
		for huc in l_curs:
			ind += 1
			print "%s..." % ind, # print how many we've done!
			if huc.HUC_12 == args:
				huc.setValue("network_distance", 0)
			else:
				distance = network_get_distance(args, huc.HUC_12,
				                                zones_info) # we could save a TON of time on this by caching the path for args[0]
				huc.setValue("network_distance", distance)
			l_curs.updateRow(huc)
	else: # just do these two hucs, setting the attribute on the new huc
		distance = network_get_distance(args[0], args[1], zones_info)
		l_curs = arcpy.UpdateCursor(zones_layer, "\"HUC_12\" = '%s'" % args[1])
		for huc in l_curs:
			huc.setValue("network_distance", distance)
			l_curs.updateRow(huc)

	return zones_layer


def network_get_distance(zone1, zone2, zones_info):
	huc1_path = []
	huc2_path = []

	network_get_path(zone=zone1, zones_downstream=zones_info, path_list=huc1_path)
	network_get_path(zone=zone2, zones_downstream=zones_info, path_list=huc2_path)

	distance = 0
	for index in range(len(huc1_path)):
		if huc1_path[index] in huc2_path: # figure out where they meet first
			distance = index
			huc2_index = huc2_path.index(huc1_path[index])
			if not huc2_index == -1:
				return distance + huc2_index
			else:
				return 20000 # it's not actually there, which also shouldn't happen

	# 20000 is a number large enough to not ever appear naturally - we can exclude it in symbology
	return 20000 # if we get here, there was a problem, OR they aren't in the same basin (so, one is klamath and the other Sac, etc


def network_get_path(zone, zones_downstream, path_list):
	global network_end_hucs

	current_DS = zone
	while (current_DS is not None):

		if current_DS in network_end_hucs:
			break
		path_list.append(current_DS)
		try:
			current_DS = zones_downstream[current_DS]
		except KeyError: # we have hucs that reference downstream hucs, but because we clip to CA, they are missing - we need to tolerate that
			break


def find_upstream(watershed, all_watersheds, dams_flag=False):
	if len(all_watersheds[watershed].upstream) > 0: # if we've already run for this watershed
		return all_watersheds[watershed].upstream

	if dams_flag and all_watersheds[watershed].has_dam:
		return [] # if this is a dam and we're account for that, return nothing upstream - can't go further

	all_us = []
	for wat in all_watersheds:
		if all_watersheds[wat].downstream == watershed:
			us = find_upstream(wat, all_watersheds, dams_flag)
			all_us.append(wat)
			all_us += us

	all_watersheds[watershed].upstream = all_us
	return all_us


def get_diversity_into_network(dams_flag, db_cursor, native_flag, qc_flag, upstream_flag, zones_layer):
	all_watersheds = {}
	# some setup

	log.write("Setting up watersheds", 1)
	reader = arcpy.SearchCursor(zones_layer)
	for record in reader:
		t_ws = watershed()
		t_ws.HUC_12 = record.HUC_12
		t_ws.downstream = record.HU_12_DS
		all_watersheds[record.HUC_12] = t_ws
	del reader
	del record

	log.write("Getting upstream hucs - this may take some time", 1)

	for l_wat in all_watersheds.keys():
		if upstream_flag == "simple":
			try:
				all_watersheds[all_watersheds[l_wat].downstream].upstream.append(
					l_wat) # add this huc to the list of upstream hucs for the one downstream
			except:
				pass # it just doesn't exist
		elif upstream_flag == "trace":
			# the following function has a return value...untested code...
			all_watersheds[all_watersheds[l_wat].downstream].upstream += find_upstream(l_wat, all_watersheds, dams_flag)

	if dams_flag:
		log.write("getting dam information", 1)

		# get the dam info
		sql = "select Zone from Zones_Aux where rim_dam = True"
		results = db_cursor.execute(sql)
		for lz in results:
			all_watersheds[lz.Zone].has_dam = True
		del results

	non_native_query = "select distinct Species_ID from Observations where Presence_Type in (%s) and Zone_ID = ?" % (
		local_vars.current_obs_types)
	native_query = "select distinct Observations.Species_ID from Observations, Fish_Species where Observations.Zone_ID = ? and Observations.Presence_Type in (%s) and Fish_Species.FID = Observations.Species_ID and Fish_Species.Native = True" % (
		local_vars.current_obs_types)
	qc_ext = "and Observations.OBJECTID = Observation_Collections.Observation_ID and Observation_Collections.Collection_ID in (%s)" % local_vars.hq_collections
	non_native_qc_query = "select distinct Species_ID from Observations, Observation_Collections where Presence_Type in (%s) and Zone_ID = ? %s" % (
		local_vars.current_obs_types, qc_ext)
	native_qc_query = "select distinct Observations.Species_ID from Observations, Fish_Species, Observation_Collections where Observations.Zone_ID = ? and Observations.Presence_Type in (%s) and Fish_Species.FID = Observations.Species_ID and Fish_Species.Native = True %s" % (
		local_vars.current_obs_types, qc_ext)

	log.write("getting assemblage information", 1)
	# get the assemblage data
	if native_flag:
		if qc_flag:
			query = native_qc_query
		else:
			query = native_query
	else:
		if qc_flag:
			query = non_native_qc_query
		else:
			query = non_native_query

	for l_wat in all_watersheds.keys():
		results = db_cursor.execute(query, l_wat)

		for species in results:  # add all the species to the assemblage
			all_watersheds[l_wat].assemblage.append(species.Species_ID)

	return all_watersheds


def diversity(zones_layer, db_cursor, args, map_layer):
	"""
		Generates 4 different layers based upon args
		args = native/non,dams/no,divtype,qc/noqc
	"""

	try:
		if args[0] == "native":
			native_flag = True
		else:
			native_flag = False
	except:
		native_flag = True

	try:
		if args[1] == "dams":
			dams_flag = True
		else:
			dams_flag = False
	except:
		dams_flag = False

	try:
		div_func = args[2]
	except:
		div_func = "diversity_jaccard"

	try:
		if args[3] == "qc":
			qc_flag = True
		else:
			qc_flag = False
	except:
		qc_flag = False

	try:
		if args[4] == "upstream_simple":
			upstream_flag = "simple"
		else:
			upstream_flag = "trace"
	except:
		upstream_flag = "simple"

	all_watersheds = get_diversity_into_network(dams_flag, db_cursor, native_flag, qc_flag, upstream_flag, zones_layer)

	log.write("calculating diversity", 1)
	# calculate the diversity

	diversity_func = globals()[div_func]

	for l_wat in all_watersheds.keys():
		all_watersheds[l_wat].beta_div = diversity_func(all_watersheds[l_wat], all_watersheds)

	log.write("updating data", 1)

	if dams_flag and native_flag:
		sql = "update Zones_Aux set beta_div_nat_cur = ? where Zone = ?"
		layer_name = "div_native_dams"
		assemblage_field_name = "assemblage_native_dams"
	elif native_flag:
		sql = "update Zones_Aux set beta_div_nat_hist = ? where Zone = ?"
		layer_name = "div_native"
		assemblage_field_name = "assemblage_native"
	elif dams_flag:
		sql = "update Zones_Aux set beta_div_nn_cur = ? where Zone = ?"
		layer_name = "div_all_dams"
		assemblage_field_name = "assemblage_all_dams"
	else:
		sql = "update Zones_Aux set beta_div_nn_hist = ? where Zone = ?"
		layer_name = "div_all"
		assemblage_field_name = "assemblage_all"

	if qc_flag:
		layer_name = "%s_qc" % layer_name
		assemblage_field_name = "%s_qc" % assemblage_field_name

	for l_wat in all_watersheds.keys():
		db_cursor.execute(sql, all_watersheds[l_wat].beta_div, l_wat)

	new_layer = arcpy.CreateUniqueName(layer_name, local_vars.workspace)
	layer_name = os.path.split(new_layer)[1]

	arcpy.env.overwriteOutput = True

	arcpy.CopyFeatures_management(zones_layer, new_layer)  # copy it out
	arcpy.MakeFeatureLayer_management(new_layer,
	                                  zones_layer)  # copy it back - need to do this to not modify original hucs

	arcpy.AddField_management(zones_layer, layer_name, "DOUBLE")
	arcpy.AddField_management(zones_layer, assemblage_field_name, "TEXT", "", "", 10000)
	updater = arcpy.UpdateCursor(zones_layer)
	for row in updater:
		cname_assemblage = diversity_assemblage_to_cname(all_watersheds[row.HUC_12].assemblage)
		#try:
		#	t_set = None
		#if type(all_watersheds[row.HUC_12].beta_div) == "int" or type(all_watersheds[row.HUC_12].beta_div) == "float":
		t_set = float(all_watersheds[row.HUC_12].beta_div)
		#except:
		#	t_set = None
		#log.write("layer_name: [%s], t_set: [%s], assemblage: [%s]" % (layer_name,t_set,str(cname_assemblage)))
		row.setValue(layer_name, t_set)
		row.setValue(assemblage_field_name, str(cname_assemblage))
		updater.updateRow(row)

	del updater

	return zones_layer


def diversity_assemblage_to_cname(assemblage):
	output_assemblage = []
	for species in assemblage:
		species = unicode(species)
		if local_vars.all_fish.has_key(species):
			output_assemblage.append(local_vars.all_fish[species].species) # append the common name
		else:
			log.error("Species in observations, but not recorded in all_fish - species code %s" % species)
	return output_assemblage


def diversity_count(current_watershed, all_watersheds):
	diversity = 0
	if current_watershed.downstream is None:
		return 0

	# if downstream is a network ending, then set the assemblage to 0
	if current_watershed.downstream in network_end_hucs:
		return 0

	for species in current_watershed.assemblage:
		try:
			if species not in all_watersheds[current_watershed.downstream].assemblage:
				diversity += 1
			for l_up in current_watershed.upstream:  # check upstream too
				if species not in all_watersheds[l_up].assemblage:
					diversity += 1
		except:
			diversity = 0

	return diversity


def diversity_jaccard(current_watershed, all_watersheds): # should already have the assemblages
	'''returns Jaccard DISSIMILARITY'''
	connected_assemblage = [] # set it here in case the next step fails
	try:
		if current_watershed.downstream not in network_end_hucs: #if it's not a network end huc
			connected_assemblage += all_watersheds[current_watershed.downstream].assemblage
	except KeyError:
		pass  # no downstream huc existing

	for l_up in current_watershed.upstream: # check upstream too
		connected_assemblage += all_watersheds[l_up].assemblage # add those to the connected assemblage
	connected_assemblage = list(set(connected_assemblage)) # remove duplicates
	assem_union = list(set(connected_assemblage) | set(current_watershed.assemblage))
	assem_intersect = list(set(connected_assemblage) & set(current_watershed.assemblage))

	len_assem_intersect = len(assem_intersect)
	len_assem_union = len(assem_union)

	if len_assem_union != 0: # don't divide by zero - the universe doesn't like that
		return (1 - (float(len_assem_intersect) / float(len_assem_union))) # return jaccard diversity (intersect/union)
	else:
		return 0;


def diversity_difference():
	pass


def forest_listing(zones_layer, db_cursor, args): # map callback
	pass