import _winreg
import shutil
import os
import sys
import string
import re

import arcpy
import pyodbc

import local_vars
import log
import input_filters


def db_connect(db_name, note=None):

	log_string = "Making database connection to %s" % db_name

	if note is not None:
		log_string += " - note added: %s" % note

	log.write(log_string)
	conn = pyodbc.connect('DRIVER={Microsoft Access Driver (*.mdb)};DBQ='+db_name)
	return conn.cursor(), conn


def db_close(cursor, conn):
	log.write("Closing database connection")
	cursor.close()
	conn.close()


def refresh_cursor(database=None, connection=None, existing_cursor=None, close_connection=False):
	if not database and not connection:
		raise ValueError("at least one of connection or database parameter is required")

	if existing_cursor:  # close a cursor if we passed one in
		existing_cursor.close()

	if close_connection:  # if the user specified to close the connection
		connection.close()
		connection = None

	if not connection:  # connection takes priority over database, so use it if we have it
		cursor, connection = db_connect(database, "refreshing cursor")
		return connection, cursor

	return connection, connection.cursor()  # always returns connection, otherwise we'd leak the handle


def compose_query_from_dict(table, value_pairs):
	query_fields = "("
	query_values = "("

	query = "INSERT INTO %s " % table

	for key in value_pairs:
		query_fields += key + ", "
		if value_pairs[key] is None:
			query_values += "NULL, "
		elif type(value_pairs[key]) is int or type(value_pairs[key]) is float:
			query_values += "%s, " % value_pairs[key]
		else:
			escape_me = str(value_pairs[key])
			escape_me = escape_me.replace("'", "''")  # replace single quotes in strings with a double single because we're going to wrap in singles.
			query_values += "'%s', " % escape_me

	query_fields = query_fields[:-2]
	query_values = query_values[:-2]  # chop off the final comma and space

	query_fields += ")"
	query_values += ")"

	query += query_fields + " VALUES " + query_values
	return query


def match_species_from_string(species_string):
	l_fid_match = re.search("(\w{3}\d{2})", species_string)
	return l_fid_match.group(0)


def create_collection(name, short_name, description, db_cursor):
	query = "insert into defs_Collections (Collection_Name, Short_Name, Description) values ('%s', '%s', ?)" % (name, short_name)
	db_cursor.execute(query, description)

	return get_last_insert_id(db_cursor)


def get_last_insert_id(db_cursor):
	"""
	A shortcut function to return the ID of the last insert operation.
	:param db_cursor: The pyodbc cursor that the insert occurred on.
	:return: integer ID of last record
	"""

	l_query = "select @@identity as id_value"
	l_identity = db_cursor.execute(l_query).fetchone()
	return l_identity.id_value


def query_to_list_of_dicts(query, db_cursor, drop_fields=None):

	records = []

	results = db_cursor.execute(query)
	fields = [t[0] for t in db_cursor.description]  # from the pyodbc website, a way to get all the fields from a cursor

	for result in results:
		record = {}
		for field in fields:
			if drop_fields and field in drop_fields:  # if we're supposed to drop the pkeys, and this field was identified as a pkey above
				continue  # go to the next iteration

			record[str(field).lower()] = result.__getattribute__(field)  # using field for both keys because we wrote them as "as" statements in the querhy
		records.append(record)

	return records


def data_stats(db_cursor):

	print "\nCalculating data stats"

	l_queries = {}
	l_queries['Total Number of Observations'] = "select Count (*) as num from observations"
	l_queries['Total Number of Datasets Included'] = "select Count(*) as num from (select distinct Set_ID from Observations)" # we use this method of gathering this because some sets have been completely tossed.
	l_queries['Total Number of Species tracked (including data bins)'] = "select Count (*) as num from Species"
	l_queries['Total Number of Species tracked (no data bins)'] = "select Count (*) as num from Species where Temporary = FALSE"
	l_queries['Total Number of Valid Obervations in the Quality Controlled Set'] = "select Count (*) as num from observation_collections where Collection_ID in (%s)" % local_vars.hq_collections
	l_queries['Total Number of Fish Species with Data'] = "select Count(*) as num from (select distinct Species_ID from Observations)"
	l_queries['Total Number of *Native* Fish Species with Data'] = "select Count(*) as num from (select distinct Observations.Species_ID from Observations, Species where Observations.Species_ID = Species.FID and Species.Native = True)"
	l_queries['Total Number of Species with Present QC Data'] = 'SELECT count(*) as num FROM (select distinct Fish_Species.FID FROM Observations, Fish_Species, Observation_Collections WHERE Observations.Species_ID = Fish_Species.FID and Observations.Presence_Type in (%s) and Observations.OBJECTID = Observation_Collections.Observation_ID and Observation_Collections.Collection_ID in (%s));' % (local_vars.current_obs_types, local_vars.hq_collections)
	l_queries['Total Number of Species with Historic QC Data'] = 'SELECT count(*) as num FROM (select distinct Fish_Species.FID FROM Observations, Fish_Species, Observation_Collections WHERE Observations.Species_ID = Fish_Species.FID and Observations.Presence_Type in (%s) and Observations.OBJECTID = Observation_Collections.Observation_ID and Observation_Collections.Collection_ID in (%s));' % (local_vars.historic_obs_types, local_vars.hq_collections)
	l_queries['Total Number of NonNative Species'] = 'Select count(*) as num from Species where Native=FALSE'

	for name in l_queries.keys():
		results = db_cursor.execute(l_queries[name])
		for result in results:
			print "%s: %s" % (name, result.num)

	print ""  # just want a newline
	# pop the info for each fish into a numpy array (numpy.array([0,1,2...]) and then print out numpy.sum() numpy.std() numpy.mean() for the natives


def new_data_get_current_observation_sets(l_connection): #get_current_data
	l_sql = 'SELECT Source_Data FROM Observation_Sets;'
	l_cursor, l_conn = db_connect(local_vars.maindb)

	l_connection.append(l_cursor)
	l_connection.append(l_conn)

	l_data = l_cursor.execute(l_sql)

	return l_data


def new_data_fetch():
	log.write("\nChecking for new data in %s" % local_vars.newdb,1)

	new_data_pull() # fetches the rows from the database that correspond to the feature sets we're about to fetch
	new_data_validate()
	new_data_dedupe() # dedupe also sets up the observation set for each item if it's not a duplicate


def new_data_validate():
	log.write("Ensuring new data rows have a corresponding dataset")

	arcpy.env.workspace = local_vars.newdb

	for new_data in local_vars.input_rows.keys(): # if a record in the database leads to a feature class that doesn't exist, remove it from the list to process
		if not arcpy.Exists(new_data):
			log.error("Skipping %s - No feature class exists for database row" % new_data)
			del local_vars.input_rows[new_data]
	else:
		log.write("All database records have a corresponding dataset")

	arcpy.env.workspace = local_vars.workspace #set the workspace back!


def new_data_dedupe(): # simply checks them against existing observations to warn if we are duplicating
	log.write("Checking new datasets for duplicates")
	l_connection = [] # this construction is stupid - it should be changed at some point, but not right now.

	l_existing_data = new_data_get_current_observation_sets(l_connection) # Check for redundant features - l_connection passed in so the db connection can be placed in it

	for item in local_vars.input_rows.keys(): #for each item in the list of features
		for existing_data in l_existing_data:
			if(existing_data == item): # check it against the list of existing data in the db
				log.write("Feature class %s is the same as %s - if you believe them to be different, please rename the new feature class" % (local_vars.input_rows[item],existing_data),1)
				del local_vars.input_rows[item]               # if if already exists, remove it from the array
				break                                   # and go to the next iteration

		print "New Feature Class: %s\n" % item

		l_dataset = local_vars.observation_set(os.path.join(local_vars.newdb,item),item) #__init__ takes the dataset path as an optional argument
		local_vars.datasets.append(l_dataset)

	db_close(l_connection[0],l_connection[1])
	del l_connection #clean up - this program could run a while and we want it clean

	#### Returning
	arcpy.env.workspace = local_vars.workspace
		#switch the workspace back


def new_data_pull():
	log.write("Checking %s for new data" % local_vars.newdb)
	l_cursor,l_conn = db_connect(local_vars.newdb)
	l_sql = 'SELECT ID,Feature_Class_Name,Species_ID,Input_Filter,Presence_Type,IF_Method,Observer_ID,Survey_Method,Notes,Data_Source_Notes,Input_Filter_Args, Projection FROM NewData where Imported = 0 order by Species_ID asc;'

	l_data = l_cursor.execute(l_sql)
	for row in l_data:
		local_vars.input_rows[row.Feature_Class_Name] = local_vars.input_data(row.ID,row.Species_ID,row.Input_Filter,
															  row.Presence_Type,row.IF_Method,
															  row.Observer_ID,row.Survey_Method,row.Notes,row.Data_Source_Notes,
															  row.Input_Filter_Args,row.Projection)
		log.write("Found data %s in input database" % row.Feature_Class_Name)

	db_close(l_cursor,l_conn)


def setup_new_data(obs_sets):
	for key in range(len(obs_sets)): #doing this with the index values so that we can delete it if we need to.
		try:
			obs_sets[key].setup(key) #passing in key so that errors can be handled in the function
		except local_vars.DataProcessingError as error_msg:
			obs_sets[key] = None # if we get an error processing the data while setting it up, remove it from the array to skip it, but then continue - it won't be processed in the future.
							# we use = None instead of del so that the array doesn't get reindexed.
							# TODO: This solution doesn't address the problem of previously committed data.
			log.write(error_msg, 1)


def get_species_from_alt_code(l_alt_code_species, filter_code):

	l_alt_code_species = string.upper(l_alt_code_species) # standardize everything coming in - they alt codes are indexed in caps

	if input_filters.alt_codes_by_filter.has_key(filter_code) and input_filters.alt_codes_by_filter[filter_code].has_key(l_alt_code_species): # if the input filter has alt codes and this alt code is defined
		return input_filters.alt_codes_by_filter[filter_code][l_alt_code_species]
	else:
		return False


def copy_data(feature_class, new_location): # get a full path and a place to copy it to, defaulting to the observations database

	l_file_split = os.path.split(feature_class)
	out_name = l_file_split[1]

	inc = 0
	return_flag = 0
	# TODO: The following code should be replaced with arcpy.CreateUniqueName
	while arcpy.Exists(os.path.join(new_location,out_name)): # if the name already exists in the output location, start renaming it with an iterator until we reach one that doesn't
		inc = inc + 1
		out_name = "%s_%02i" %(out_name[:-3],inc)
		return_flag = 1

	arcpy.Copy_management(feature_class, os.path.join(new_location, out_name))
	log.write("Saving %s to %s" % (feature_class, new_location))

	if return_flag == 1:  # we need to send back the new name
		return out_name


def clean_workspace():  # gets a list of all of the feature classes in the workspace and removes them all

	log.write("Cleaning workspace - %s" % local_vars.workspace)

	if arcpy.Exists(local_vars.workspace):  # if we have a workspace, delete it
		arcpy.Delete_management(local_vars.workspace)
		#if os.path.isdir(local_vars.workspace):  # using this method because arcpy.Delete_management wasn't deleting it
		#	shutil.rmtree(local_vars.workspace)
		#else:
		#	os.remove(local_vars.workspace)

	workspace_parts = os.path.split(local_vars.workspace)
	arcpy.CreateFileGDB_management(workspace_parts[0], workspace_parts[1]) # then recreate it - fastest way to compact a temp GDB


def setup_test_mode():  # test mode lets us simulate the script, but copies the datafiles to new ones and uses those
	#                    this lets us simulate an import before conducting it, etc

	import string
	import shutil
	log.write(auto_print=1, log_string="Running in test mode. Copying data files - this may take some time...")

	#make the path for test mode

	try:
		test_path = local_vars.test_folder
		log.write("Checking for test folder and replacing old test files")

		if not os.path.exists(test_path): # if the test folder doesn't exist
			os.makedirs(test_path) # make it!

	except:
		log.write(log_string="Unable to start in test mode - couldn't make the test directory - exiting", auto_print=1)
		raise

	#copy dbs and set new path
	l_folders = ["data","inputs","log","maps","mxds","proc"]
	l_exclude = ["Hillshade","data_storage","maps","output","temp.gdb"] # having maps here too will create the folder structure but not move the data
	for index in range(len(l_folders)): # make a listing of the main folders we want to copy - this new method will be slower since it only checks the mod time on the folders and not the files
		l_folders[index] = os.path.join(local_vars.internal_workspace,l_folders[index])

	# if this ends up being slow, it would probably be relatively quick for us to iterate over the files IN each of the folders we name above (if it has files, in this case they all do)
	# and then check for modified data there before copying


	for folder in l_folders:
		m_folder = string.replace(folder,local_vars.internal_workspace,test_path)
		if not os.path.exists(m_folder):
			os.mkdir(m_folder)

		try:  #before copying files, see if it has been modified. If it has, remove the old one and add the new one.
			l_data = os.walk(folder)

			for i_folder in l_data: # this will iterate over every folder, subfolder, etc

				folder_base = i_folder[0]
				new_folder_base = folder_base

				new_folder_base = string.replace(new_folder_base,local_vars.internal_workspace,test_path)
				if new_folder_base[-1:] == "\\": # chop off the trailing slash
					new_folder_base = new_folder_base[:-1]

				for subfolder in i_folder[1]: # folder[1] will hold this folder's own subfolder - make the folders
					new_folder = os.path.join(new_folder_base,subfolder)
					if not os.path.exists(new_folder):
						os.mkdir(new_folder)

				continue_flag = 0 # skip this folder if it's in our skip list, but only AFTER we've created the subfolders - we always want the tree, but we might not want the items.
				for exclude_item in l_exclude: # skip these items though
					if not (re.search(exclude_item,i_folder[0]) == None):
						continue_flag = 1

					#print i_folder[0]
				if continue_flag == 1: # structured this way to break out of the outer loop
					continue


				for file in i_folder[2]: # copy all the files over
					old_path = os.path.join(folder_base,file)
					new_path = os.path.join(new_folder_base,file)

					l_mtime1 = int(os.path.getmtime(old_path)) # get the modified time

					l_new_file_exists = os.path.exists(new_path) #rather than call it twice, store the result
					if l_new_file_exists:
						l_mtime2 = int(os.path.getmtime(new_path))
					else: # it doesn't exist yet
						l_mtime2 = 0 # set it, but it won't be equal, so it'll copy in the next step

					if l_new_file_exists and l_mtime1 == l_mtime2 : # if it already exists in the test folder AND the files were modified at the same time
						pass #they're the same! Awesome - don't do anything
					else: #otherwise, we need to change some things
						if l_new_file_exists:
							os.remove(new_path) # same story - out with the old
						shutil.copy2(old_path, new_path)  # create a new one

		except:
			log.write("Unable to start in test mode - couldn't copy the folder %s - exiting" % folder, 1)
			raise # raise this exception up

	local_vars.set_workspace_vars(test_path)  # resets them all with test_path as the base


def remove_data(item_name, l_workspace=arcpy.env.workspace):  # checks if a feature exists, then deletes it
	t_wkspc_save = arcpy.env.workspace  # @UndefinedVariable

	arcpy.env.workspace = l_workspace  # get the workspace of the function calling this so we remove the correct item
	# this should actually be carried across functions...should use the same workspace as before.

	if arcpy.Exists(item_name):  # if it exists
		arcpy.Delete_management(item_name)  # delete it

	arcpy.env.workspace = t_wkspc_save


def is_between(check_num, num1, num2): # checks whether check_num is between num1 and num2 without assumptions of the sign of num1 or num2 since a max in ArcGIS could be in a different quadrant from the min

	if num2 < check_num < num1:
		return True
	if num2 > check_num > num1:
		return True

	return False


def get_path():

	"""
	Obtains the PISCES install path from the windows system registry - used for scripts running in ArcGIS where
	the current working directory isn't the PISCES install path.

	:return: str base_folder
	:raise: WindowsError on failure to retrive info from the registry
	"""
	try:
		registry = _winreg.ConnectRegistry("",_winreg.HKEY_CURRENT_USER) # open the registry
		base_folder = _winreg.QueryValue(registry,"Software\CWS\PISCES\location") # get the PISCES location
		_winreg.CloseKey(registry)
	except:
		raise WindowsError("Unable to get base folder")

	return base_folder


def text_to_species_list(text_input):

	if not local_vars.data_setup_run:
		local_vars.data_setup()

	text_input = str(text_input)

	# Check if it's just a fish - return it as a list
	if text_input in local_vars.all_fish:
		return [text_input]

	# Ok, maybe then it's a group - try it first as the full group name
	group_species = species_group_as_list(text_input)
	if group_species:
		return group_species

	# then try it as the short_name
	group_species = species_group_as_list(short_name = text_input)
	if group_species:
		return group_species

	# if we're still here, then what was passed in was NOT a single group and not a single fish
	# Check if it's a comma or ; separated list of items. Split and validate each item is a fish, then return
	temp_items = []
	if text_input.find(",") > 0:
		temp_items = text_input.split(",")
	elif text_input.find(";") > 0:
		temp_items = text_input.split(";")

	final_list = []
	if len(temp_items) > 0:
		for item in temp_items:
			# is it a fish?
			if item in local_vars.all_fish:
				final_list.append(item)
			else:
				# maybe it's a group
				t_group = text_to_species_list(item) # maybe it's a group or sort of string - we could also just try the two group checks again
				if len(t_group) > 0:
					final_list += t_group
		return list(set(final_list)) # it could be empty - dedupe too

	# still here? We've got nothing. Return [] - still valid for many of these items, and won't raise an error. Let them check themselves
	log.error("Text contains no reference to species")
	return []

def species_group_as_list(group_name = None,short_name=None,db_cursor = None):

	if group_name:
		select_group_id = "select ID from defs_Species_groups where Group_Name='%s'" % group_name
	elif short_name:
		select_group_id = "select ID from defs_Species_groups where Short_Name='%s'" % short_name
	else:
		log.error("No group specified")
		return None

	if not db_cursor:
		db_cursor,db_conn = db_connect(local_vars.maindb)

	result = db_cursor.execute(select_group_id)
	id_row = result.fetchone()
	try:
		group_id = id_row.ID
	except:
		return None

	del id_row, result

	if not group_id:
		return None

	group_query = "select FID from Species_Groups where Group_ID = ?"
	results = db_cursor.execute(group_query,group_id)

	species = []
	for row in results:
		species.append(row.FID)

	db_close(db_cursor, db_conn)

	return species


def hucs_to_list(dataset):

	'''takes an arcgis dataset containing a field with "HUC_12" in it and transforms the unique values in that field into a list'''

	# check that we have a dataset
	if not dataset:
		log.error("No HUCs to process")
		return False

	# and check that it has a huc field
	desc = arcpy.Describe(dataset)
	for field in desc.fields:
		# probably better done with a lambda
		if local_vars.huc_field == field.name:
			break
	else: 
		log.error("dataset has no HUC_12 field - can't process")
		return False

	# cleanup
	del desc

	huc_list = []

	# get the list
	huc_cursor = arcpy.SearchCursor(dataset)
	for row in huc_cursor:
		huc_list.append(row.getValue(local_vars.huc_field))

	# dedupe it on the way out
	return list(set(huc_list))


def refresh_layer_cache(l_db_cursor=None):
	"""
		deletes the existing layer cache and recreates it
		:param layer_cache_location: relative path to layer cache
	"""

	if not l_db_cursor:
		l_db_cursor, l_db_conn = db_connect(local_vars.maindb, "Cleaning Layer Cache")

	layer_cache = local_vars.layer_cache

	log.write("Deleting layer cache - will recreate", 1)

	if os.path.exists(layer_cache):  # if it currently exists
		arcpy.Delete_management(layer_cache)

	arcpy.CreateFileGDB_management(os.path.split(layer_cache)[0], os.path.split(layer_cache)[1], "CURRENT")

	query = "delete * from Layer_Cache"  # delete the records too!
	l_db_cursor.execute(query)

	if l_db_conn:
		l_db_cursor.close()
		l_db_conn.close()


def get_species_in_dataset(dataset_id, db_cursor):
	query = "select distinct Species_ID from Observations, Observation_Collections where Observations.OBJECTID = Observation_Collections.Observation_ID and Observation_Collections.Collection_ID = %s" % dataset_id
	results = db_cursor.execute(query)

	species_list = []
	for record in results:
		species_list.append(record.Species_ID)

	return species_list

def run_query(query):
	"""
		A simple helper function that helps run a query with less fuss
	"""

	db_cursor, db_conn = db_connect(local_vars.maindb, "run_query")

	return db_cursor, db_conn, db_cursor.execute(query)


