import os
import sys
import traceback
from _winreg import *

import arcpy
import pyodbc

from PISCES import __version__

# workspace variables

debug = False
version = __version__

internal_workspace = None  # essentially the pisces root folder
maindb = None
newdb = None
workspace = None
calcs_mdb = None
temp = None
observationsdb = None
layer_cache = None
geo_aux = None

HUCS = None
huc_field = "HUC_12"
mxd_source = None
mxd_ddp_source = None
test_folder = None
auto_project = None

all_fish = None
input_rows = None
datasets = None

# Flags set in the arguments that tell it to skip mapping or importing respectively
importonly = 0
maponly = 0
usecache = 0  # skip generating new map layers, use the cache where possible

arc_path = None
projections_path = None
proj_teale_albers = None
proj_utm = None
projections = None  # it's ok that it's None now - it will update when the initial variables do.
default_proj = None

# utilities
seven_zip = None  # where 7zip is located

# license files
data_license = None

data_setup_run = False  # flag to let us know if this has run already so we don't run it twice

set_workspace_run = False

layer_cache_blank = "blank_feature"

# tables
zones_aux = "Zones_Aux"
zones_table = "HUC12FullState"
observations_table = "Observations"
species_table = "Species"
fish_species_table = "Fish_Species"
species_aux_table = "Species_Aux"
maps_table = "defs_Query_Sets"
bind_vals_table = "Query_Bind"
reg_key = "Software\CWS\PISCES\location"

tracker_url = "https://bitbucket.org/nickrsan/pisces/issues"

current_obs_types = "1,3,6,7,9"
notrans_obs_types = "1,3,9"  # a collection that includes only non-translocated populations. We can switch current to it with a command line flag
historic_obs_types = "2,5,10"
hq_collections = "5"


def set_workspace_vars(l_workspace=None):
	global internal_workspace, maindb, newdb, workspace, calcs_mdb, temp, observationsdb, layer_cache, geo_aux
	global HUCS, mxd_source, mxd_ddp_source, test_folder
	global auto_project
	global all_fish, input_rows, datasets
	global arc_path, projections_path, proj_teale_albers, proj_utm, projections, default_proj
	global seven_zip
	global data_license
	global set_workspace_run
    
	if l_workspace is None and not set_workspace_run:  # if no l_workspace is provided, then do this to set the workspace

		try:
			try:
				registry = ConnectRegistry("", HKEY_CURRENT_USER)  # open the registry
				base_folder = QueryValue(registry, reg_key)  # get the PISCES location
				CloseKey(registry)
			except:
				log.error("Unable to get base folder")
			
			os.chdir(base_folder)
			internal_workspace = base_folder
			set_workspace_run = True
			sys.path.append(base_folder)
		except:
			log.write("Unable to set workspace - you are likely loading PISCES from another module. You must manually run set_workspace_vars if you wish to do this", 1)
		
	else:  # set it to what we've provided - this is most likely to occur when called from the test setup function to reset all paths
		internal_workspace = l_workspace
	
	maindb = os.path.join(internal_workspace, "data", "pisces.mdb")
	newdb = os.path.join(internal_workspace, "inputs", "new_data.mdb")
	workspace = os.path.join(internal_workspace, "proc", "calculations.gdb")
		# TODO: this var may need to be adjusted since so many arc scripts just use it intrinsically for input.
	calcs_mdb = os.path.join(internal_workspace, "proc", "calculations.mdb")
	temp = os.path.join(internal_workspace, "proc", "temp")
	observationsdb = os.path.join(internal_workspace, "data", "observations.gdb")
	layer_cache = os.path.join(internal_workspace, "data", "layer_cache.gdb")  # TODO: Make argument
	geo_aux = os.path.join(internal_workspace, "data", "geo_aux.mdb")  # TODO: Make argument
	HUCS = os.path.join(maindb, "HUCs", zones_table)
	
	mxd_source = os.path.join(internal_workspace, "mxds", "base", "blank.mxd")
	mxd_ddp_source = os.path.join(internal_workspace, "mxds", "base", "blank_ddp.mxd") # a default to use for USFS layers - probably not particularly useful to have since we're specifiying that doc anyway
	
	test_folder = os.path.join(internal_workspace, "test") #folder where data goes if we are in test mode
	
	#OPTIONS
	auto_project = 1 # sets whether we should attempt to reproject datasets that aren't in Teale Albers. If this is 1, then we will. TODO: Make parameter
	
	# data variables
	all_fish = {}
	input_rows = {} # stores all of the input data by filename for access at the appropriate time - avoids extra SQL connections/releases and can be accessed later
	datasets = []

	# ArcGIS folder
	arc_path = arcpy.GetInstallInfo()['InstallDir']  # won't work with the environment variable, and nothing in the registry seems to contain the base Arc folder.
	projections_path = os.path.join(arc_path, "Coordinate Systems")
	proj_teale_albers = os.path.join(projections_path, "Projected Coordinate Systems", "State Systems", "NAD 1983 California (Teale) Albers (Meters).prj")
	proj_utm = os.path.join(projections_path,"Projected Coordinate Systems","UTM","NAD 1983","NAD 1983 UTM Zone 10N.prj")
	projections = {'UTM':proj_utm, 'Teale_Albers':proj_teale_albers}
	default_proj = proj_teale_albers
	
	seven_zip = os.path.join(internal_workspace,"utils","7za.exe")
	data_license = os.path.join(internal_workspace,"license.htm")

set_workspace_vars()  # call it immediately

# import other PISCES items after running set_workspace_vars
import funcs
import log
import input_filters
import observations

# Error Classes

class GenError(Exception):
	def __init__(self, value):
		self.value = value
		log.error(value)
	def __str__(self):
		return repr(self.value)

class IFError(Exception): #an IFError relates to the input filter and determining it. We want to catch it separately
	def __init__(self, value):
		self.value = value
		log.error(str(value))
	def __str__(self):
		return repr(self.value)

class DataProcessingError(Exception):
	def __init__(self, value = None,set_index = None):
		
		if not set_index == None: # if the dataset is provided when the error is raised, remove the dataset from processing
			datasets[set_index] = None #using = None instead of del so that it doesn't get reindexed. We'll need to check for a value when we loop now.
		
		log.error(str(value))
		
		self.value = value # set the error string
	def __str__(self):
		return "\nCritical Error:" + repr(self.value)


class DataStorageError(Exception):  # used after all the data is processed to detect custom errors

	def __init__(self, value=None):
	
		self.value = value  # set the error string
		log.error(value)

	def __str__(self):
		return "\nCritical Error:" + repr(self.value)


class MappingError(Exception):  # used after all the data is processed to detect custom errors

	def __init__(self, value=None):
	
		self.value = value  # set the error string
		
		log.error("%s - stack trace follows: %s" % (value, traceback.print_exc()))
		
	def __str__(self):
		return "\nMapping Error:" + repr(self.value)
	
# Main program classes


class fish:  # this class will be used for internal program representation of the fish data table for quick local lookups
	def __init__(self):
		self.species = None
		self.fid = None
		self.alt_codes = []
		
		self.sci_name = None
		#a dictionary of this class will be maintained based on the fid for lookup

class input_data: # used for temporary storage of the db rows

	# TODO: change i1,i2, etc in __init__ below to be more semantic
	def __init__(self, i1=None, i2=None, i3=None, i4=None, i5=None, i6=None, i7=None, i8=None, i9=None, i10=None, i11=None):  # be ready for all of it to be passed in via constructor
		self.ID = i1
		self.Species_ID = i2
		self.Input_Filter = i3
		self.Presence_Type = i4
		self.IF_Method = i5
		self.Observer_ID = i6
		self.Survey_Method = i7
		self.Notes = i8
		self.Source_Data_Notes = i9
		self.Input_Filter_Args = i10
		self.Projection_Key = i11


class observation:  # contains data related to a particular observation data file
	def __init__(self):
		self.observer = None
		self.species_id = None
		self.zone_id = None  # Zone == HUC - trying to make this program a little more generic
		self.observation_date = None
		self.presence_type = None
		self.latitude = None  # TODO: should this be the polygon centroid for polygons? Same for next
		self.longitude = None
		self.notes = None  # info from the input filter, etc that gets tacked on to the records
		self.other_data = None  # place to store additional data from an import that we think is valuable to keep in a human readable form
		self.certainty = None
		self.survey_method = None
		self.collections = []  # added 12/13/2011 - input filters can add collection_ids to this list and any collections present will be inserted for the observation
		self.objectid = None  # added 12/13/2011 - lets us set this here and call another method on the db to insert collections
		self.set_id = None  # added 10/10/2013 - does not reference parent object on imports. Just an integer, and not set by importing code
		self.if_method = None  # same as previous
		self.date_added = None  # similar to previous. Not set by Input Filter, but used when selecting observations back out

		self.table_used = "Observations"  # used when selecting from other tables that are like observations

		self.not_db_items = ('not_db_items', 'table_used', 'collections', 'do_not_add', 'observer', 'table_specific_not_db_items')  # items that don't correspond to the obs table - observer is because it's a property of if_method, really
		self.do_not_add = ('objectid',)  # items to not include in an insert
		self.table_specific_not_db_items = {'Invalid_Observations': ['date_added', ], 'Observations': []}  # similar to not_db, but for specific tables. a kluge. TODO: Make this a better solution

	def load(self, record_id, from_table=None, db_cursor=None):
		"""
			Given an observation id, loads the relevant observation from the database into this object
			use from_table to override the table to use (ie, to make an Invalid_Observations object. If db_cursor
			is provided, it will be used for the database. Otherwise, a new connection will be made.
		"""
		log.write("Objectid: %s" % record_id, 1)
		if from_table:
			self.table_used = from_table

		close_cursor = False
		db_conn = None
		if not db_cursor:  # open a new cursor if we don't have one yet, note in var that we opened it so we can close it
			db_cursor, db_conn = funcs.db_connect(maindb, "loading observations")
			close_cursor = True

		select_str = "SELECT OBJECTID, Set_ID, Species_ID, zone_id, notes, Longitude, Latitude, " \
					 "Presence_Type, IF_Method, Certainty, Other_Data, Survey_Method, Observation_Date"
		if self.table_used.lower() == "observations":  # hack to make it work slightly differently for the two separate tables
			select_str += ", Date_Added"
		select_str += " FROM %s where OBJECTID = ?" % self.table_used

		results = db_cursor.execute(select_str, record_id)  # pass record_id as bind value
		record = results.fetchone()
		if not record:
			return

		fields = [t[0] for t in db_cursor.description]
		for field in fields:
			self.__dict__[str(field).lower()] = record.__getattribute__(field)  # using field for both keys because we wrote them as "as" statements in the querhy

		# ok, now get the collections and append them to the list on the object
		select_collections = "SELECT Collection_ID from Observation_Collections where Observation_ID = ?"
		collections_res = db_cursor.execute(select_collections, record_id)
		for collection in collections_res:
			self.collections.append(collection.Collection_ID)

		if close_cursor:
			funcs.db_close(db_cursor, db_conn)

	def insert(self, db_cursor):
		"""
			If we have data, then insert it into the database as a new record, returning the record id and setting it as an attribute. Use builtin funcs for this, passing a dictionary

		:param db_cursor:
		"""

		# The following is a funky way of copying the object dictionary while excluding a few items (the items that don't
		# correspond to DB fields) - creates a new dict, looks at every key on the object, and exclude the fields defined
		# in the exclusion variables not_db_items and do_not_add

		insert_dict = {}
		for key in self.__dict__:
			if key not in self.not_db_items and key not in self.do_not_add and key not in self.table_specific_not_db_items[self.table_used]:
				insert_dict[key] = self.__dict__[key]

		query = funcs.compose_query_from_dict(self.table_used, insert_dict)
		db_cursor.execute(query)

		# get the objectid
		l_query = "select @@identity as id_value"
		l_identity = db_cursor.execute(l_query).fetchone()
		self.objectid = l_identity.id_value

		self.insert_collections(db_cursor)  # now insert the collections for this record

	def insert_collections(self, db_cursor):
		"""

		:param db_cursor:
		:return:
		"""

		if self.table_used.lower() != "observations":
			log.write("skipping collections addition because table is %s" % self.table_used)
			return

		for collection_id in self.collections:
			query = "insert into observation_collections (observation_id, collection_id) values (%s,%s)" % (self.objectid, collection_id)

			try:
				db_cursor.execute(query)
			except pyodbc.IntegrityError as err:
				# likely means that this record already exists if we're violating an integrity constraint
				log.write("skipping insert for record %s and collection_id %s because it failed a constraint"
						  " (IntegrityError) - probably a duplicate" % (self.objectid, collection_id))


class result_set:
	def __init__(self, p_observations=[], p_if_method=None):
		self.observations = p_observations  # contains observation objects
		self.if_method = p_if_method


class observation_set:  # basic data structure that holds the data we'll be adding to the database
	
	def __init__(self, l_dataset_path=None, l_dataset_name=None):  # take it optionally, but allow it to be set later
		self.results = []  # contains result_set objects
		self.species_fid = None
		self.observer = None
		self.obs_date = None
		self.obs_type = None
		self.survey_method = None
		self.dataset_path = l_dataset_path 
		self.dataset_name = l_dataset_name
		self.dataset_key = None  # where am I?? Lets it know where it can find itself in the datasets so that it can delete itself
		self.filter_code = None  #the extracted filter defining code from the filename
		self.input_filter = None  #the actual, initialized input filter
		self.set_id = None  # in the database, that is
		self.new_data = None  # the new data object for this item will be moved here
		
		log.write("New observation set constructed - not yet added to the database")

	def setup(self, dataset_key):

		log.write("Setting up observation set %s" % self.dataset_name)

		self.new_data = input_rows[self.dataset_name]
		del input_rows[self.dataset_name]

		# Set the values to what was brought in with self.new_data
		self.filter_code = self.new_data.Input_Filter
		self.species_fid = self.new_data.Species_ID
		self.observer = self.new_data.Observer_ID
		self.dataset_key = dataset_key
		self.survey_method = self.new_data.Survey_Method

		# Check input filter - make sure it's valid. If it isn't skip the file
		try:
			self.make_input_filter()
			self.check_species()
		except DataProcessingError as error_msg:
			log.write(error_msg, 1)  # it already should have already removed itself from the list, so long as it was raised with the index value
			return  # and stop setting this one up
					
		# TODO: Validate ALL other params. Some can be blank, but if they ARE defined, they need to be valid.
		
		self.new_db_obs_set()
	
	def make_input_filter(self):
		
		#first, check if we have one
		if self.filter_code == None:
			raise DataProcessingError("No input filter set!",self.dataset_key) # if we don't have an input filter, we can stop processing right here
		
		#then make sure it's a valid input filter - could be combined with above, but let's keep it separate for readability
		if not input_filters.IF_List.has_key(self.filter_code):
			raise DataProcessingError("Input filter is invalid. Check the database to make sure it exists - you provided %s" % self.filter_code,self.dataset_key) # if the specified filter doesn't exist, we can stop processing right here
		
		#then, check if we have code for it??
		try:
			l_class = getattr(input_filters, input_filters.IF_List[self.filter_code]) #get the actual class object into l_class
			self.input_filter = l_class(self) # then make a new input filter - the IF_List maps the code to the class - we call it with self as an argument so that it knows where its observation set is
		except:
			raise DataProcessingError("Unable to create input filter object for filter code %s - class should be %s - check to make sure there is a class for the filter code" % (self.filter_code, input_filters.IF_List[self.filter_code]),self.dataset_key)
			
	def check_species(self):
		
		if self.species_fid == "filter":
			#TODO: This should test if the function exists. For tables, we'll need to have determine_species pass back MUL99 (Multiple species - it matches the other format for validation, but is clearly out of the ordinary)
			self.input_filter.determine_species()
		elif all_fish.has_key(self.species_fid):
			pass #we're good! It's already set correctly
		else: #if we've still got nothing, it's either an alt_code or gibberish - if it's not an alt_code, then the following function will throw a DataProcessingError
			self.species_fid = funcs.get_species_from_alt_code(self.species_fid,self.filter_code)
		
		if self.species_fid == "filter" or self.species_fid == None or self.species_fid is False: # the previous attempts failed...
			raise DataProcessingError("Unable to determine the species ID for dataset %s" % self.dataset_name,self.dataset_key)
		
	def process_data(self):
		self.input_filter.process_observations()  # dumps results as observations into self.results[]
		# TODO: The following line is a temporary fix - it shouldn't be there and should also do actual processing to determine what should be in place of the "1" - the input filter should retrieve that
		
		#self.add_observations()
		
		#self.import_to_db()
		
		# self.cleanup()
		
	def new_db_obs_set(self):
		
		log.write("Creating new observation set for data")
		
		l_cursor, l_connection = funcs.db_connect(maindb)
		
		try:
			l_query = "insert into Observation_Sets (Species,Input_Filter,Observer,Notes,Dataset_Notes) values (?,?,?,?,?)"
			l_cursor.execute(l_query,self.species_fid,self.filter_code,self.observer,self.new_data.Notes,self.new_data.Source_Data_Notes)
			l_connection.commit()
		except:
			raise DataProcessingError("Unable to add a record to table Observation_Sets for %s. Skipping data - this dataset will be automatically retried next time you run the program")
		
		l_query = "select @@identity as id_value"
		
		l_identity = l_cursor.execute(l_query)
		
		for item in l_identity:
			self.set_id = item.id_value
		log.data_write("Observation Set %s created for %s" % (self.set_id, self.dataset_name))
		
		funcs.db_close(l_cursor, l_connection)
	
	def insert_data(self, db_cursor):  # eventually, once all the processing is done, this will add it all to the db
		# TODO: Test this function on an import. Refactoring didn't catch some renames here
		# TODO: Refactor this to use the observation.insert() method

		l_sql = "insert into observations (Set_ID,Species_ID,Zone_ID,Presence_Type,IF_Method,Certainty,Longitude,Latitude,Notes,Other_Data,Observation_Date,Survey_Method) values (?,?,?,?,?,?,?,?,?,?,?,?)"
		for result in self.results:
			for obs in result.observations:
				db_cursor.execute(l_sql, self.set_id, obs.species_id, obs.zone_id, obs.presence_type, result.if_method.method_id, obs.certainty, obs.longitude,obs.latitude,obs.notes,obs.other_data, obs.observation_date,obs.survey_method)
				
				# this will slow things down a bit, but probably marginally in human time...
				id_result = db_cursor.execute("select @@IDENTITY as t_id")
				for id_row in id_result:
					obs.db_id = id_row.t_id				
	
	def insert_collections(self, db_cursor):
		l_sql = "insert into observation_collections (Observation_ID, Collection_ID) values (?,?)"
		for result in self.results:
			for obs in result.observations:
				for c_id in obs.collections:  # basically, if there are any collections, insert them
					db_cursor.execute(l_sql, obs.db_id, c_id)
	
	def record_data(self, db_cursor):
		log.write("Inserting new records for %s" % self.dataset_name, 1)
		log.data_write("Inserting new records for %s" % self.dataset_name)
		self.insert_data(db_cursor)  # insert the processed observations into the database
		
		self.insert_collections(db_cursor)  # the objectid is already set on all new observations
		
		log.write("Copying data source %s to observations database" % self.dataset_name,1)
		self.save_data(db_cursor)  # copies the source datafiles to the observations storage database and updates the source locations of the data
	
	def save_data(self, db_cursor):  # once we've added the observations, this copies the dataset to the observations database and updates the path in the object and the database
		
		#copy features
		
		new_name = funcs.copy_data(self.dataset_path,observationsdb) #the location defaults to the observations database
			# it returns a new name in case it has to be renamed to be copied
		
		#delete old features from new db
		
		funcs.remove_data(self.dataset_name, newdb)
		if not new_name == None: # if the copy function modified the dataset name
			self.dataset_name = new_name

		#update the database and the object with the new dataset location
		
		self.dataset_path = os.path.join(observationsdb,self.dataset_name)
		l_sql = "update Observation_Sets set Source_Data = ? where Set_ID = ?"
		db_cursor.execute(l_sql,self.dataset_path,self.set_id)

		#mark the input row as imported - we want that info saved in case something went wrong, so don't delete it
		self.cleanup_mark_input_row_imported()
		
		#TODO: Fix import and update code

	def cleanup_mark_input_row_imported(self):
		l_cursor,l_conn = funcs.db_connect(newdb)
		sql_query = "update NewData set Imported = ? where ID = ?"
		l_cursor.execute(sql_query,self.set_id,self.new_data.ID) #TODO change input_rows[...] to the objects path part
			# setting imported = set_id so that we can reimport directly from this record and the stored data in the future
		
		l_conn.commit()
		funcs.db_close(l_cursor, l_conn)


def fully_qualified_path(relative_item):
	"""
		takes a path relative to the PISCES root and returns its full path. Somewhat trivial, but complicated when the
		fully PISCES code isn't running. Does no validation of inputs to ensure it's a true path
	"""

	if not internal_workspace:
		set_workspace_vars()  # run the first setup if this code hasn't been run yet

	return os.path.join(internal_workspace, relative_item)

def print_workspace_vars():
	print "\n\nUsing workspace:"
	print "\tMain Database: %s" % maindb
	print "\tInternal Workspace: %s" % internal_workspace
	print "\tWorkspace: %s" % workspace
	print "\tNew Data Database: %s" % newdb
	print "\tObservations Database: %s" % observationsdb
	print "\tLayer Cache: %s" % layer_cache
	print "\tAuxiliary Geographic Data: %s\n" % geo_aux
	print "\tUsing Observation Types in: %s\n" % current_obs_types
	print "\tUsing Historic Observation Types in: %s\n" % historic_obs_types
	print "\tUsing Collections in (when using high quality collections): %s\n" % hq_collections
	print ""
	
	log.write("Using main database %s,<br/> workspace %s,<br/> new data location %s" % (maindb, arcpy.env.workspace, newdb)) #@UndefinedVariable


def data_setup():
	"""
		Main function calling various others to set up internal variables that speed up PISCES
	"""
					
	log.write("Loading Configuration Data", 1)
	get_species_data()
	observations.check_observations_db()  # checks to make sure we have an observations db to import to
	input_filters.get_IF_list()  # populate the IF list
	input_filters.retrieve_if_methods()  # populate the listing of methods for each IF
	input_filters.pull_alt_codes()
	observations.get_observation_certainties()  # populates the dictionary of default certainties for observations
	pull_field_maps()

	global data_setup_run
	data_setup_run = True


def initialize():
	"""
		Just an alias for data_setup
	:return:
	"""

	data_setup()


def get_species_data():
	
	log.write("Retrieving species data from main database", 1)

	db_cursor, db_conn = funcs.db_connect(maindb)
	
	l_sql = 'select FID,Common_Name, Scientific_Name from Species'
	l_data = db_cursor.execute(l_sql)

	global all_fish
	
	for l_item in l_data:  # for every fish that's returned from the database
		all_fish[l_item.FID] = fish()  # create a new fish object in the all_fish dictionary, keyed on the fid
		all_fish[l_item.FID].fid = l_item.FID  # and add the parameter data
		all_fish[l_item.FID].species = l_item.Common_Name
		all_fish[l_item.FID].sci_name = l_item.Scientific_Name
	
	funcs.db_close(db_cursor, db_conn)


def pull_field_maps():
	
	db_cursor,db_conn = funcs.db_connect(newdb, "Getting field maps")
	
	l_sql = "select f1.NewData_ID, f1.Field_Name, f1.Input_Field, f1.Handler_Function, f1.Required from FieldMapping as f1 LEFT OUTER JOIN NewData as f2 ON f1.NewData_ID = f2.ID where (f2.Imported = 0 or f1.NewData_ID = 0) order by f1.NewData_ID asc" # "where Imported = 0" might not work well...but it should - that's the criteria on the importing data as wellselect f1.NewData_ID, f1.Field_Name, f1.Input_Field, f1.Handler_Function, f1.Required from FieldMapping as f1 OUTER JOIN NewData as f2 ON f1.NewData_ID = f2.ID where f2.Imported = 0 order by f1.NewData_ID asc" # "where Imported = 0" might not work well...but it should - that's the criteria on the importing data as wellselect f1.NewData_ID, f1.Field_Name, f1.Input_Field, f1.Handler_Function, f1.Required from FieldMapping as f1 OUTER JOIN NewData as f2 ON f1.NewData_ID = f2.ID where f2.Imported = 0 order by f1.NewData_ID asc" # "where Imported = 0" might not work well...but it should - that's the criteria on the importing data as well
	rows = db_cursor.execute(l_sql)
	
	for row in rows:
		if not input_filters.field_maps.has_key(row.NewData_ID): # if we don't already have the array in place
			input_filters.field_maps[row.NewData_ID] = [] # make it into an empty array
			
		l_map = input_filters.field_map()
		l_map.field_name = row.Field_Name
		l_map.handler_function = row.Handler_Function
		l_map.input_field = row.Input_Field
		l_map.set_id = row.NewData_ID
		l_map.required = row.Required
		
		input_filters.field_maps[row.NewData_ID].append(l_map)
		
	funcs.db_close(db_cursor, db_conn)


def projection_code_to_file(projection_code):  # probably doesn't need to be a function (since it's just looking things up in a dict), but in the interest of a standard interface, this works
	return projections[projection_code]


def get_version():
	"""
		Just a compatibility function
	:return:
	"""
	return version